@@ -65,7 +65,7 @@
 struct dhcp_opt {
 	uint32_t option; /* Also used for IANA Enterpise Number */
 	int type;
-	int len;
+	size_t len;
 	char *var;
 
 	int index; /* Index counter for many instances of the same option */
@@ -80,7 +80,7 @@ struct dhcp_opt {
 	size_t encopts_len;
 };
 
-struct dhcp_opt *vivso_find(uint16_t, const void *);
+struct dhcp_opt *vivso_find(uint32_t, const void *);
 
 size_t dhcp_vendor(char *, size_t);
 
@@ -91,16 +91,17 @@ int make_option_mask(const struct dhcp_opt *, size_t,
     uint8_t *, const char *, int);
 
 size_t encode_rfc1035(const char *src, uint8_t *dst);
-ssize_t decode_rfc3397(char *, ssize_t, int, const uint8_t *);
-ssize_t print_string(char *, ssize_t, int, const uint8_t *);
-ssize_t print_option(char *, ssize_t, int, int, const uint8_t *, const char *);
+ssize_t decode_rfc3397(char *, size_t, const uint8_t *, size_t);
+ssize_t print_string(char *, size_t, const uint8_t *, size_t);
+ssize_t print_option(char *, size_t, int, const uint8_t *, size_t,
+    const char *);
 
-ssize_t dhcp_envoption(struct dhcpcd_ctx *,
+size_t dhcp_envoption(struct dhcpcd_ctx *,
     char **, const char *, const char *, struct dhcp_opt *,
     const uint8_t *(*dgetopt)(struct dhcpcd_ctx *,
-    unsigned int *, unsigned int *, unsigned int *,
-    const uint8_t *, unsigned int, struct dhcp_opt **),
-    const uint8_t *od, int ol);
+    size_t *, unsigned int *, size_t *,
+    const uint8_t *, size_t, struct dhcp_opt **),
+    const uint8_t *od, size_t ol);
 void dhcp_zero_index(struct dhcp_opt *);
 
 #endif
