@@ -35,8 +35,9 @@
 #endif
 
 #include <arpa/inet.h>
+#include <net/if.h>
 #include <net/route.h>
-
+#include <netinet/if_ether.h>
 #include <netinet/in_systm.h>
 #include <netinet/in.h>
 #include <netinet/ip.h>
@@ -62,6 +63,7 @@
 #include "dhcp-common.h"
 #include "duid.h"
 #include "eloop.h"
+#include "if.h"
 #include "ipv4.h"
 #include "ipv4ll.h"
 #include "script.h"
@@ -80,6 +82,10 @@
 #define RELEASE_DELAY_S		0
 #define RELEASE_DELAY_NS	10000000
 
+#ifndef IPDEFTTL
+#define IPDEFTTL 64 /* RFC1340 */
+#endif
+
 struct dhcp_op {
 	uint8_t value;
 	const char *name;
@@ -135,7 +141,7 @@ dhcp_printoptions(const struct dhcpcd_ctx *ctx)
 #define get_option_raw(ctx, dhcp, opt) get_option(ctx, dhcp, opt, NULL)
 static const uint8_t *
 get_option(struct dhcpcd_ctx *ctx,
-    const struct dhcp_message *dhcp, uint8_t opt, int *len)
+    const struct dhcp_message *dhcp, unsigned int opt, size_t *len)
 {
 	const uint8_t *p = dhcp->options;
 	const uint8_t *e = p + sizeof(dhcp->options);
@@ -144,7 +150,7 @@ get_option(struct dhcpcd_ctx *ctx,
 	uint8_t overl = 0;
 	uint8_t *bp = NULL;
 	const uint8_t *op = NULL;
-	int bl = 0;
+	size_t bl = 0;
 
 	while (p < e) {
 		o = *p++;
@@ -216,7 +222,7 @@ get_option_addr(struct dhcpcd_ctx *ctx,
     uint8_t option)
 {
 	const uint8_t *p;
-	int len;
+	size_t len;
 
 	p = get_option(ctx, dhcp, option, &len);
 	if (!p || len < (ssize_t)sizeof(a->s_addr))
@@ -230,7 +236,7 @@ get_option_uint32(struct dhcpcd_ctx *ctx,
     uint32_t *i, const struct dhcp_message *dhcp, uint8_t option)
 {
 	const uint8_t *p;
-	int len;
+	size_t len;
 	uint32_t d;
 
 	p = get_option(ctx, dhcp, option, &len);
@@ -247,7 +253,7 @@ get_option_uint8(struct dhcpcd_ctx *ctx,
     uint8_t *i, const struct dhcp_message *dhcp, uint8_t option)
 {
 	const uint8_t *p;
-	int len;
+	size_t len;
 
 	p = get_option(ctx, dhcp, option, &len);
 	if (!p || len < (ssize_t)sizeof(*p))
@@ -258,10 +264,11 @@ get_option_uint8(struct dhcpcd_ctx *ctx,
 }
 
 ssize_t
-decode_rfc3442(char *out, ssize_t len, int pl, const uint8_t *p)
+decode_rfc3442(char *out, size_t len, const uint8_t *p, size_t pl)
 {
 	const uint8_t *e;
-	ssize_t b, bytes = 0, ocets;
+	size_t bytes = 0, ocets;
+	int b;
 	uint8_t cidr;
 	struct in_addr addr;
 	char *o = out;
@@ -302,23 +309,23 @@ decode_rfc3442(char *out, ssize_t len, int pl, const uint8_t *p)
 		} else
 			b = snprintf(o, len, "0.0.0.0/0");
 		o += b;
-		len -= b;
+		len -= (size_t)b;
 
 		/* Finally, snag the router */
 		memcpy(&addr.s_addr, p, 4);
 		p += 4;
 		b = snprintf(o, len, " %s", inet_ntoa(addr));
 		o += b;
-		len -= b;
+		len -= (size_t)b;
 	}
 
 	if (out)
 		return o - out;
-	return bytes;
+	return (ssize_t)bytes;
 }
 
 static struct rt_head *
-decode_rfc3442_rt(int dl, const uint8_t *data)
+decode_rfc3442_rt(const uint8_t *data, size_t dl)
 {
 	const uint8_t *p = data;
 	const uint8_t *e;
@@ -366,10 +373,11 @@ decode_rfc3442_rt(int dl, const uint8_t *data)
 }
 
 char *
-decode_rfc3361(int dl, const uint8_t *data)
+decode_rfc3361(const uint8_t *data, size_t dl)
 {
 	uint8_t enc;
-	unsigned int l;
+	size_t l;
+	ssize_t r;
 	char *sip = NULL;
 	struct in_addr addr;
 	char *p;
@@ -383,11 +391,12 @@ decode_rfc3361(int dl, const uint8_t *data)
 	dl--;
 	switch (enc) {
 	case 0:
-		if ((l = decode_rfc3397(NULL, 0, dl, data)) > 0) {
+		if ((r = decode_rfc3397(NULL, 0, data, dl)) > 0) {
+			l = (size_t)r;
 			sip = malloc(l);
 			if (sip == NULL)
 				return 0;
-			decode_rfc3397(sip, l, dl, data);
+			decode_rfc3397(sip, l, data, dl);
 		}
 		break;
 	case 1:
@@ -403,7 +412,8 @@ decode_rfc3361(int dl, const uint8_t *data)
 		while (dl != 0) {
 			memcpy(&addr.s_addr, data, sizeof(addr.s_addr));
 			data += sizeof(addr.s_addr);
-			p += snprintf(p, l - (p - sip), "%s ", inet_ntoa(addr));
+			p += snprintf(p, l - (size_t)(p - sip),
+			    "%s ", inet_ntoa(addr));
 			dl -= sizeof(addr.s_addr);
 		}
 		*--p = '\0';
@@ -420,7 +430,7 @@ decode_rfc3361(int dl, const uint8_t *data)
  * separated string. Returns length of string (including
  * terminating zero) or zero on error. */
 ssize_t
-decode_rfc5969(char *out, ssize_t len, int pl, const uint8_t *p)
+decode_rfc5969(char *out, size_t len, const uint8_t *p, size_t pl)
 {
 	uint8_t ipv4masklen, ipv6prefixlen;
 	uint8_t ipv6prefix[16];
@@ -456,7 +466,7 @@ decode_rfc5969(char *out, ssize_t len, int pl, const uint8_t *p)
 		    ipv6prefix[12],ipv6prefix[13],ipv6prefix[14], ipv6prefix[15]
 		);
 
-		len -= b;
+		len -= (size_t)b;
 		out += b;
 		bytes += b;
 	} else {
@@ -473,7 +483,7 @@ decode_rfc5969(char *out, ssize_t len, int pl, const uint8_t *p)
 		if (out) {
 			b= snprintf(out, len, " %d.%d.%d.%d",
 			    br[0], br[1], br[2], br[3]);
-			len -= b;
+			len -= (size_t)b;
 			out += b;
 			bytes += b;
 		} else {
@@ -488,7 +498,7 @@ static char *
 get_option_string(struct dhcpcd_ctx *ctx,
     const struct dhcp_message *dhcp, uint8_t option)
 {
-	int len;
+	size_t len;
 	const uint8_t *p;
 	char *s;
 
@@ -544,7 +554,7 @@ get_option_routes(struct interface *ifp, const struct dhcp_message *dhcp)
 	const uint8_t *e;
 	struct rt_head *routes = NULL;
 	struct rt *route = NULL;
-	int len;
+	size_t len;
 	const char *csr = "";
 
 	/* If we have CSR's then we MUST use these only */
@@ -559,7 +569,7 @@ get_option_routes(struct interface *ifp, const struct dhcp_message *dhcp)
 			csr = "MS ";
 	}
 	if (p) {
-		routes = decode_rfc3442_rt(len, p);
+		routes = decode_rfc3442_rt(p, len);
 		if (routes) {
 			if (!(ifo->options & DHCPCD_CSR_WARNED)) {
 				syslog(LOG_DEBUG,
@@ -644,7 +654,7 @@ dhcp_message_add_addr(struct dhcp_message *dhcp,
 		p += *p + 1;
 	}
 
-	len = p - (uint8_t *)dhcp;
+	len = (size_t)(p - (uint8_t *)dhcp);
 	if (len + 6 > sizeof(*dhcp)) {
 		errno = ENOMEM;
 		return -1;
@@ -665,8 +675,7 @@ make_message(struct dhcp_message **message,
 	uint8_t *n_params = NULL;
 	uint32_t ul;
 	uint16_t sz;
-	size_t len, i;
-	int auth_len;
+	size_t len, i, auth_len;
 	const struct dhcp_opt *opt;
 	struct if_options *ifo = iface->options;
 	const struct dhcp_state *state = D_CSTATE(iface);
@@ -695,11 +704,11 @@ make_message(struct dhcp_message **message,
 	}
 
 	dhcp->op = DHCP_BOOTREQUEST;
-	dhcp->hwtype = iface->family;
+	dhcp->hwtype = (uint8_t)iface->family;
 	switch (iface->family) {
 	case ARPHRD_ETHER:
 	case ARPHRD_IEEE802:
-		dhcp->hwlen = iface->hwlen;
+		dhcp->hwlen = (uint8_t)iface->hwlen;
 		memcpy(&dhcp->chaddr, &iface->hwaddr, iface->hwlen);
 		break;
 	}
@@ -748,7 +757,7 @@ make_message(struct dhcp_message **message,
 	if (type == DHCP_DECLINE) {
 		*p++ = DHO_MESSAGE;
 		len = strlen(DAD);
-		*p++ = len;
+		*p++ = (uint8_t)len;
 		memcpy(p, DAD, len);
 		p += len;
 	}
@@ -776,19 +785,21 @@ make_message(struct dhcp_message **message,
 	    type == DHCP_INFORM ||
 	    type == DHCP_REQUEST)
 	{
+		int mtu;
+
 		*p++ = DHO_MAXMESSAGESIZE;
 		*p++ = 2;
-		sz = get_mtu(iface->name);
-		if (sz < MTU_MIN) {
-			if (set_mtu(iface->name, MTU_MIN) == 0)
+		mtu = if_getmtu(iface->name);
+		if (mtu < MTU_MIN) {
+			if (if_setmtu(iface->name, MTU_MIN) == 0)
 				sz = MTU_MIN;
-		} else if (sz > MTU_MAX) {
+		} else if (mtu > MTU_MAX) {
 			/* Even though our MTU could be greater than
 			 * MTU_MAX (1500) dhcpcd does not presently
 			 * handle DHCP packets any bigger. */
-			sz = MTU_MAX;
+			mtu = MTU_MAX;
 		}
-		sz = htons(sz);
+		sz = htons(mtu);
 		memcpy(p, &sz, 2);
 		p += 2;
 
@@ -821,6 +832,15 @@ make_message(struct dhcp_message **message,
 			    ifo->options & DHCPCD_HOSTNAME_SHORT ? 1 : 0);
 		else
 			hostname = ifo->hostname;
+
+		/*
+		 * RFC4702 3.1 States that if we send the Client FQDN option
+		 * then we MUST NOT also send the Host Name option.
+		 * Technically we could, but that is not RFC conformant and
+		 * also seems to break some DHCP server implemetations such as
+		 * Windows. On the other hand, ISC dhcpd is just as non RFC
+		 * conformant by not accepting a partially qualified FQDN.
+		 */
 		if (ifo->fqdn != FQDN_DISABLE) {
 			/* IETF DHC-FQDN option (81), RFC4702 */
 			*p++ = DHO_FQDN;
@@ -843,14 +863,14 @@ make_message(struct dhcp_message **message,
 			*p++ = 0; /* from server for PTR RR */
 			*p++ = 0; /* from server for A RR if S=1 */
 			if (hostname) {
-				ul = encode_rfc1035(hostname, p);
-				*lp += ul;
-				p += ul;
+				i = encode_rfc1035(hostname, p);
+				*lp += i;
+				p += i;
 			}
 		} else if (ifo->options & DHCPCD_HOSTNAME && hostname) {
 			*p++ = DHO_HOSTNAME;
 			len = strlen(hostname);
-			*p++ = len;
+			*p++ = (uint8_t)len;
 			memcpy(p, hostname, len);
 			p += len;
 		}
@@ -882,7 +902,7 @@ make_message(struct dhcp_message **message,
 			    i < ifo->vivco_len;
 			    i++, vivco++)
 			{
-				len = (p - m) + vivco->len + 1;
+				len = (size_t)(p - m) + vivco->len + 1;
 				if (len > sizeof(*dhcp))
 					goto toobig;
 				if (vivco->len + 2 + *lp > 255) {
@@ -899,7 +919,7 @@ make_message(struct dhcp_message **message,
 			}
 		}
 
-		len = (p - m) + 3;
+		len = (size_t)((p - m) + 3);
 		if (len > sizeof(*dhcp))
 			goto toobig;
 		*p++ = DHO_PARAMETERREQUESTLIST;
@@ -918,12 +938,12 @@ make_message(struct dhcp_message **message,
 			    (opt->option == DHO_RENEWALTIME ||
 				opt->option == DHO_REBINDTIME))
 				continue;
-			len = (p - m) + 2;
+			len = (size_t)((p - m) + 2);
 			if (len > sizeof(*dhcp))
 				goto toobig;
-			*p++ = opt->option;
+			*p++ = (uint8_t)opt->option;
 		}
-		*n_params = p - n_params - 1;
+		*n_params = (uint8_t)(p - n_params - 1);
 	}
 
 	/* silence GCC */
@@ -931,17 +951,18 @@ make_message(struct dhcp_message **message,
 	auth = NULL;
 
 	if (ifo->auth.options & DHCPCD_AUTH_SEND) {
-		auth_len = dhcp_auth_encode(&ifo->auth, state->auth.token,
+		auth_len = (size_t)dhcp_auth_encode(&ifo->auth,
+		    state->auth.token,
 		    NULL, 0, 4, type, NULL, 0);
 		if (auth_len > 0) {
-			len = (p + auth_len) - m;
+			len = (size_t)((p + auth_len) - m);
 			if (auth_len > 255 || len > sizeof(*dhcp))
 				goto toobig;
 			*p++ = DHO_AUTHENTICATION;
 			*p++ = (uint8_t)auth_len;
 			auth = p;
 			p += auth_len;
-		} else if (auth_len == -1)
+		} else if ((ssize_t)auth_len == -1)
 			syslog(LOG_ERR, "%s: dhcp_auth_encode: %m",
 			    iface->name);
 	}
@@ -956,13 +977,13 @@ make_message(struct dhcp_message **message,
 		*p++ = DHO_PAD;
 #endif
 
-	len = p - m;
+	len = (size_t)(p - m);
 	if (ifo->auth.options & DHCPCD_AUTH_SEND && auth_len > 0)
 		dhcp_auth_encode(&ifo->auth, state->auth.token,
 		    m, len, 4, type, auth, auth_len);
 
 	*message = dhcp;
-	return len;
+	return (ssize_t)len;
 
 toobig:
 	syslog(LOG_ERR, "%s: DHCP messge too big", iface->name);
@@ -974,6 +995,7 @@ ssize_t
 write_lease(const struct interface *ifp, const struct dhcp_message *dhcp)
 {
 	int fd;
+	size_t len;
 	ssize_t bytes;
 	const uint8_t *e, *p;
 	uint8_t l;
@@ -996,11 +1018,11 @@ write_lease(const struct interface *ifp, const struct dhcp_message *dhcp)
 	/* Only write as much as we need */
 	p = dhcp->options;
 	e = p + sizeof(dhcp->options);
-	bytes = sizeof(*dhcp);
+	len = sizeof(*dhcp);
 	while (p < e) {
 		o = *p;
 		if (o == DHO_END) {
-			bytes = p - (const uint8_t *)dhcp;
+			len = (size_t)(p - (const uint8_t *)dhcp);
 			break;
 		}
 		p++;
@@ -1009,7 +1031,7 @@ write_lease(const struct interface *ifp, const struct dhcp_message *dhcp)
 			p += l;
 		}
 	}
-	bytes = write(fd, dhcp, bytes);
+	bytes = write(fd, dhcp, len);
 	close(fd);
 	return bytes;
 }
@@ -1023,7 +1045,7 @@ read_lease(struct interface *ifp)
 	ssize_t bytes;
 	const uint8_t *auth;
 	uint8_t type;
-	int auth_len;
+	size_t auth_len;
 
 	fd = open(state->leasefile, O_RDONLY);
 	if (fd == -1) {
@@ -1073,7 +1095,7 @@ read_lease(struct interface *ifp)
 }
 
 static const struct dhcp_opt *
-dhcp_getoverride(const struct if_options *ifo, uint16_t o)
+dhcp_getoverride(const struct if_options *ifo, unsigned int o)
 {
 	size_t i;
 	const struct dhcp_opt *opt;
@@ -1090,8 +1112,8 @@ dhcp_getoverride(const struct if_options *ifo, uint16_t o)
 
 static const uint8_t *
 dhcp_getoption(struct dhcpcd_ctx *ctx,
-    unsigned int *os, unsigned int *code, unsigned int *len,
-    const uint8_t *od, unsigned int ol, struct dhcp_opt **oopt)
+    size_t *os, unsigned int *code, size_t *len,
+    const uint8_t *od, size_t ol, struct dhcp_opt **oopt)
 {
 	size_t i;
 	struct dhcp_opt *opt;
@@ -1102,8 +1124,8 @@ dhcp_getoption(struct dhcpcd_ctx *ctx,
 			return NULL;
 		}
 		*os = 2; /* code + len */
-		*code = (int)*od++;
-		*len = (int)*od++;
+		*code = (unsigned int)*od++;
+		*len = (size_t)*od++;
 		if (*len > ol) {
 			errno = EINVAL;
 			return NULL;
@@ -1126,22 +1148,21 @@ dhcp_env(char **env, const char *prefix, const struct dhcp_message *dhcp,
 {
 	const struct if_options *ifo;
 	const uint8_t *p;
-	int pl;
 	struct in_addr addr;
 	struct in_addr net;
 	struct in_addr brd;
 	struct dhcp_opt *opt, *vo;
-	ssize_t e = 0;
+	size_t e, i, pl;
 	char **ep;
 	char cidr[4];
 	uint8_t overl = 0;
-	size_t i;
 	uint32_t en;
 
+	e = 0;
 	ifo = ifp->options;
 	get_option_uint8(ifp->ctx, &overl, dhcp, DHO_OPTIONSOVERLOADED);
 
-	if (!env) {
+	if (env == NULL) {
 		if (dhcp->yiaddr || dhcp->ciaddr)
 			e += 5;
 		if (*dhcp->bootfile && !(overl & 1))
@@ -1174,7 +1195,7 @@ dhcp_env(char **env, const char *prefix, const struct dhcp_message *dhcp,
 			e += dhcp_envoption(ifp->ctx, NULL, NULL, ifp->name,
 			    opt, dhcp_getoption, p, pl);
 		}
-		return e;
+		return (ssize_t)e;
 	}
 
 	ep = env;
@@ -1309,8 +1330,8 @@ dhcp_fallback(void *arg)
 	struct interface *iface;
 
 	iface = (struct interface *)arg;
-	select_profile(iface, iface->options->fallback);
-	start_interface(iface);
+	dhcpcd_selectprofile(iface, iface->options->fallback);
+	dhcpcd_startinterface(iface);
 }
 
 uint32_t
@@ -1422,7 +1443,7 @@ eexit:
 
 static ssize_t
 dhcp_sendpacket(const struct interface *iface, struct in_addr to,
-    const uint8_t *data, ssize_t len)
+    const uint8_t *data, size_t len)
 {
 	struct sockaddr_in sin;
 
@@ -1439,6 +1460,7 @@ checksum(const void *data, uint16_t len)
 {
 	const uint8_t *addr = data;
 	uint32_t sum = 0;
+	uint16_t res;
 
 	while (len > 1) {
 		sum += addr[0] * 256 + addr[1];
@@ -1452,13 +1474,13 @@ checksum(const void *data, uint16_t len)
 	sum = (sum >> 16) + (sum & 0xffff);
 	sum += (sum >> 16);
 
-	sum = htons(sum);
+	res = htons(sum);
 
-	return ~sum;
+	return ~res;
 }
 
 static struct udp_dhcp_packet *
-dhcp_makeudppacket(ssize_t *sz, const uint8_t *data, size_t length,
+dhcp_makeudppacket(size_t *sz, const uint8_t *data, size_t length,
 	struct in_addr source, struct in_addr dest)
 {
 	struct udp_dhcp_packet *udpp;
@@ -1497,7 +1519,7 @@ dhcp_makeudppacket(ssize_t *sz, const uint8_t *data, size_t length,
 
 	ip->ip_v = IPVERSION;
 	ip->ip_hl = sizeof(*ip) >> 2;
-	ip->ip_id = arc4random() & UINT16_MAX;
+	ip->ip_id = (uint16_t)arc4random_uniform(UINT16_MAX);
 	ip->ip_ttl = IPDEFTTL;
 	ip->ip_len = htons(sizeof(*ip) + sizeof(*udp) + length);
 	ip->ip_sum = checksum(ip, sizeof(*ip));
@@ -1507,14 +1529,15 @@ dhcp_makeudppacket(ssize_t *sz, const uint8_t *data, size_t length,
 }
 
 static void
-send_message(struct interface *iface, int type,
+send_message(struct interface *iface, uint8_t type,
     void (*callback)(void *))
 {
 	struct dhcp_state *state = D_STATE(iface);
 	struct if_options *ifo = iface->options;
 	struct dhcp_message *dhcp;
 	struct udp_dhcp_packet *udp;
-	ssize_t len, r;
+	size_t len;
+	ssize_t r;
 	struct in_addr from, to;
 	in_addr_t a = 0;
 	struct timeval tv;
@@ -1531,7 +1554,8 @@ send_message(struct interface *iface, int type,
 				state->interval = 64;
 		}
 		tv.tv_sec = state->interval + DHCP_RAND_MIN;
-		tv.tv_usec = arc4random() % (DHCP_RAND_MAX_U - DHCP_RAND_MIN_U);
+		tv.tv_usec = (suseconds_t)arc4random_uniform(
+		    (DHCP_RAND_MAX - DHCP_RAND_MIN) * 1000000);
 		timernorm(&tv);
 		syslog(LOG_DEBUG,
 		    "%s: sending %s (xid 0x%x), next in %0.1f seconds",
@@ -1557,7 +1581,10 @@ send_message(struct interface *iface, int type,
 		a = state->addr.s_addr;
 		state->addr.s_addr = 0;
 	}
-	len = make_message(&dhcp, iface, type);
+	r = make_message(&dhcp, iface, type);
+	if (r == -1)
+		goto fail;
+	len = (size_t)r;
 	if (a)
 		state->addr.s_addr = a;
 	from.s_addr = dhcp->ciaddr;
@@ -1572,13 +1599,15 @@ send_message(struct interface *iface, int type,
 			dhcp_close(iface);
 		}
 	} else {
+		size_t ulen;
+
 		r = 0;
-		udp = dhcp_makeudppacket(&r, (uint8_t *)dhcp, len, from, to);
+		udp = dhcp_makeudppacket(&ulen, (uint8_t *)dhcp, len, from, to);
 		if (udp == NULL) {
 			syslog(LOG_ERR, "dhcp_makeudppacket: %m");
 		} else {
-			r = ipv4_sendrawpacket(iface, ETHERTYPE_IP,
-			    (uint8_t *)udp, r);
+			r = if_sendrawpacket(iface, ETHERTYPE_IP,
+			    (uint8_t *)udp, ulen);
 			free(udp);
 		}
 		/* If we failed to send a raw packet this normally means
@@ -1587,17 +1616,26 @@ send_message(struct interface *iface, int type,
 		 * As such we remove it from consideration without actually
 		 * stopping the interface. */
 		if (r == -1) {
-			syslog(LOG_ERR, "%s: ipv4_sendrawpacket: %m",
+			syslog(LOG_ERR, "%s: if_sendrawpacket: %m",
 			    iface->name);
-			if (!(iface->ctx->options & DHCPCD_TEST))
-				dhcp_drop(iface, "FAIL");
-			dhcp_close(iface);
-			eloop_timeout_delete(iface->ctx->eloop, NULL, iface);
-			callback = NULL;
+			switch(errno) {
+			case ENETDOWN:
+			case ENETRESET:
+			case ENETUNREACH:
+				break;
+			default:
+				if (!(iface->ctx->options & DHCPCD_TEST))
+					dhcp_drop(iface, "FAIL");
+				dhcp_close(iface);
+				eloop_timeout_delete(iface->ctx->eloop,
+				    NULL, iface);
+				callback = NULL;
+			}
 		}
 	}
 	free(dhcp);
 
+fail:
 	/* Even if we fail to send a packet we should continue as we are
 	 * as our failure timeouts will change out codepath when needed. */
 	if (callback)
@@ -1642,42 +1680,43 @@ send_rebind(void *arg)
 void
 dhcp_discover(void *arg)
 {
-	struct interface *iface = arg;
-	struct dhcp_state *state = D_STATE(iface);
-	struct if_options *ifo = iface->options;
-	int timeout = ifo->timeout;
+	struct interface *ifp = arg;
+	struct dhcp_state *state = D_STATE(ifp);
+	struct if_options *ifo = ifp->options;
+	time_t timeout = ifo->timeout;
 
 	/* If we're rebooting and we're not daemonised then we need
 	 * to shorten the normal timeout to ensure we try correctly
 	 * for a fallback or IPv4LL address. */
 	if (state->state == DHS_REBOOT &&
-	    !(iface->ctx->options & DHCPCD_DAEMONISED))
+	    !(ifp->ctx->options & DHCPCD_DAEMONISED))
 	{
-		timeout -= ifo->reboot;
-		if (timeout <= 0)
+		if (ifo->reboot >= timeout)
 			timeout = 2;
+		else
+			timeout -= ifo->reboot;
 	}
 
 	state->state = DHS_DISCOVER;
-	state->xid = dhcp_xid(iface);
-	eloop_timeout_delete(iface->ctx->eloop, NULL, iface);
+	state->xid = dhcp_xid(ifp);
+	eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
 	if (ifo->fallback)
-		eloop_timeout_add_sec(iface->ctx->eloop,
-		    timeout, dhcp_fallback, iface);
+		eloop_timeout_add_sec(ifp->ctx->eloop,
+		    timeout, dhcp_fallback, ifp);
 	else if (ifo->options & DHCPCD_IPV4LL &&
 	    !IN_LINKLOCAL(htonl(state->addr.s_addr)))
 	{
 		if (IN_LINKLOCAL(htonl(state->fail.s_addr)))
 			timeout = RATE_LIMIT_INTERVAL;
-		eloop_timeout_add_sec(iface->ctx->eloop,
-		    timeout, ipv4ll_start, iface);
+		eloop_timeout_add_sec(ifp->ctx->eloop,
+		    timeout, ipv4ll_start, ifp);
 	}
 	if (ifo->options & DHCPCD_REQUEST)
 		syslog(LOG_INFO, "%s: soliciting a DHCP lease (requesting %s)",
-		    iface->name, inet_ntoa(ifo->req_addr));
+		    ifp->name, inet_ntoa(ifo->req_addr));
 	else
-		syslog(LOG_INFO, "%s: soliciting a DHCP lease", iface->name);
-	send_discover(iface);
+		syslog(LOG_INFO, "%s: soliciting a DHCP lease", ifp->name);
+	send_discover(ifp);
 }
 
 static void
@@ -1805,18 +1844,22 @@ dhcp_bind(void *arg)
 				lease->leasetime = DHCP_MIN_LEASE;
 			}
 			if (lease->rebindtime == 0)
-				lease->rebindtime = lease->leasetime * T2;
+				lease->rebindtime =
+				    (uint32_t)(lease->leasetime * T2);
 			else if (lease->rebindtime >= lease->leasetime) {
-				lease->rebindtime = lease->leasetime * T2;
+				lease->rebindtime =
+				    (uint32_t)(lease->leasetime * T2);
 				syslog(LOG_WARNING,
 				    "%s: rebind time greater than lease "
 				    "time, forcing to %"PRIu32" seconds",
 				    iface->name, lease->rebindtime);
 			}
 			if (lease->renewaltime == 0)
-				lease->renewaltime = lease->leasetime * T1;
+				lease->renewaltime =
+				    (uint32_t)(lease->leasetime * T1);
 			else if (lease->renewaltime > lease->rebindtime) {
-				lease->renewaltime = lease->leasetime * T1;
+				lease->renewaltime =
+				    (uint32_t)(lease->leasetime * T1);
 				syslog(LOG_WARNING,
 				    "%s: renewal time greater than rebind "
 				    "time, forcing to %"PRIu32" seconds",
@@ -1850,18 +1893,18 @@ dhcp_bind(void *arg)
 		lease->renewaltime = lease->rebindtime = lease->leasetime;
 	else {
 		eloop_timeout_add_sec(iface->ctx->eloop,
-		    lease->renewaltime, dhcp_renew, iface);
+		    (time_t)lease->renewaltime, dhcp_renew, iface);
 		eloop_timeout_add_sec(iface->ctx->eloop,
-		    lease->rebindtime, dhcp_rebind, iface);
+		    (time_t)lease->rebindtime, dhcp_rebind, iface);
 		eloop_timeout_add_sec(iface->ctx->eloop,
-		    lease->leasetime, dhcp_expire, iface);
+		    (time_t)lease->leasetime, dhcp_expire, iface);
 		syslog(LOG_DEBUG,
 		    "%s: renew in %"PRIu32" seconds, rebind in %"PRIu32
 		    " seconds",
 		    iface->name, lease->renewaltime, lease->rebindtime);
 	}
 	ipv4_applyaddr(iface);
-	if (daemonise(iface->ctx) == 0) {
+	if (dhcpcd_daemonise(iface->ctx) == 0) {
 		if (!ipv4ll)
 			arp_close(iface);
 		state->state = DHS_BOUND;
@@ -1971,7 +2014,7 @@ dhcp_inform(struct interface *ifp)
 }
 
 void
-dhcp_reboot_newopts(struct interface *ifp, int oldopts)
+dhcp_reboot_newopts(struct interface *ifp, unsigned long long oldopts)
 {
 	struct if_options *ifo;
 	struct dhcp_state *state = D_STATE(ifp);
@@ -2042,8 +2085,10 @@ dhcp_reboot(struct interface *ifp)
 	/* Don't bother ARP checking as the server could NAK us first. */
 	if (ifo->options & DHCPCD_INFORM)
 		dhcp_inform(ifp);
-	else
-		dhcp_request(ifp);
+	else {
+		/* Don't call dhcp_request as that would change the state */
+		send_request(ifp);
+	}
 }
 
 void
@@ -2180,8 +2225,8 @@ dhcp_handledhcp(struct interface *iface, struct dhcp_message **dhcpp,
 	uint8_t type, tmp;
 	const uint8_t *auth;
 	struct in_addr addr;
-	size_t i;
-	int auth_len;
+	unsigned int i;
+	size_t auth_len;
 	char *msg;
 
 	/* We may have found a BOOTP server */
@@ -2327,7 +2372,8 @@ dhcp_handledhcp(struct interface *iface, struct dhcp_message **dhcpp,
 			case 1:
 				log_dhcp(LOG_WARNING, "IPv4LL enabled from",
 				    iface, dhcp, from);
-				eloop_timeout_delete(iface->ctx->eloop, NULL, iface);
+				eloop_timeout_delete(iface->ctx->eloop,
+				    NULL, iface);
 				if (IN_LINKLOCAL(htonl(state->addr.s_addr)))
 					eloop_timeout_add_sec(iface->ctx->eloop,
 					    DHCP_MAX, dhcp_discover, iface);
@@ -2347,7 +2393,7 @@ dhcp_handledhcp(struct interface *iface, struct dhcp_message **dhcpp,
 	/* Ensure that all required options are present */
 	for (i = 1; i < 255; i++) {
 		if (has_option_mask(ifo->requiremask, i) &&
-		    get_option_uint8(iface->ctx, &tmp, dhcp, i) != 0)
+		    get_option_uint8(iface->ctx, &tmp, dhcp, (uint8_t)i) != 0)
 		{
 			/* If we are bootp, then ignore the need for serverid.
 			 * To ignore bootp, require dhcp_message_type. */
@@ -2453,7 +2499,6 @@ dhcp_handledhcp(struct interface *iface, struct dhcp_message **dhcpp,
 			state->claims = 0;
 			state->probes = 0;
 			state->conflicts = 0;
-			state->state = DHS_PROBE;
 			arp_probe(iface);
 			return;
 		}
@@ -2462,7 +2507,7 @@ dhcp_handledhcp(struct interface *iface, struct dhcp_message **dhcpp,
 	dhcp_bind(iface);
 }
 
-static ssize_t
+static size_t
 get_udp_data(const uint8_t **data, const uint8_t *udp)
 {
 	struct udp_dhcp_packet p;
@@ -2526,55 +2571,58 @@ valid_udp_packet(const uint8_t *data, size_t data_len, struct in_addr *from,
 static void
 dhcp_handlepacket(void *arg)
 {
-	struct interface *iface = arg;
+	struct interface *ifp = arg;
 	struct dhcp_message *dhcp = NULL;
 	const uint8_t *pp;
-	ssize_t bytes;
+	size_t bytes;
 	struct in_addr from;
-	int i, partialcsum = 0;
-	const struct dhcp_state *state = D_CSTATE(iface);
+	int i, flags;
+	const struct dhcp_state *state = D_CSTATE(ifp);
 
-	/* We loop through until our buffer is empty.
-	 * The benefit is that if we get >1 DHCP packet in our buffer and
-	 * the first one fails for any reason, we can use the next. */
-	for(;;) {
-		bytes = ipv4_getrawpacket(iface, ETHERTYPE_IP,
-		    iface->ctx->packet, udp_dhcp_len, &partialcsum);
-		if (bytes == 0 || bytes == -1)
+	/* Need this API due to BPF */
+	flags = 0;
+	while (!(flags & RAW_EOF)) {
+		bytes = (size_t)if_readrawpacket(ifp, ETHERTYPE_IP,
+		    ifp->ctx->packet, udp_dhcp_len, &flags);
+		if (bytes == 0 || (ssize_t)bytes == -1) {
+			syslog(LOG_ERR, "%s: dhcp if_readrawpacket: %m",
+			    ifp->name);
+			dhcp_close(ifp);
 			break;
-		if (valid_udp_packet(iface->ctx->packet, bytes,
-			&from, partialcsum) == -1)
+		}
+		if (valid_udp_packet(ifp->ctx->packet, bytes,
+			&from, flags & RAW_PARTIALCSUM) == -1)
 		{
 			syslog(LOG_ERR, "%s: invalid UDP packet from %s",
-			    iface->name, inet_ntoa(from));
+			    ifp->name, inet_ntoa(from));
 			continue;
 		}
-		i = whitelisted_ip(iface->options, from.s_addr);
+		i = whitelisted_ip(ifp->options, from.s_addr);
 		if (i == 0) {
 			syslog(LOG_WARNING,
 			    "%s: non whitelisted DHCP packet from %s",
-			    iface->name, inet_ntoa(from));
+			    ifp->name, inet_ntoa(from));
 			continue;
 		} else if (i != 1 &&
-		    blacklisted_ip(iface->options, from.s_addr) == 1)
+		    blacklisted_ip(ifp->options, from.s_addr) == 1)
 		{
 			syslog(LOG_WARNING,
 			    "%s: blacklisted DHCP packet from %s",
-			    iface->name, inet_ntoa(from));
+			    ifp->name, inet_ntoa(from));
 			continue;
 		}
-		if (iface->flags & IFF_POINTOPOINT &&
+		if (ifp->flags & IFF_POINTOPOINT &&
 		    state->dst.s_addr != from.s_addr)
 		{
 			syslog(LOG_WARNING,
 			    "%s: server %s is not destination",
-			    iface->name, inet_ntoa(from));
+			    ifp->name, inet_ntoa(from));
 		}
-		bytes = get_udp_data(&pp, iface->ctx->packet);
-		if ((size_t)bytes > sizeof(*dhcp)) {
+		bytes = get_udp_data(&pp, ifp->ctx->packet);
+		if (bytes > sizeof(*dhcp)) {
 			syslog(LOG_ERR,
 			    "%s: packet greater than DHCP size from %s",
-			    iface->name, inet_ntoa(from));
+			    ifp->name, inet_ntoa(from));
 			continue;
 		}
 		if (dhcp == NULL) {
@@ -2587,22 +2635,22 @@ dhcp_handlepacket(void *arg)
 		memcpy(dhcp, pp, bytes);
 		if (dhcp->cookie != htonl(MAGIC_COOKIE)) {
 			syslog(LOG_DEBUG, "%s: bogus cookie from %s",
-			    iface->name, inet_ntoa(from));
+			    ifp->name, inet_ntoa(from));
 			continue;
 		}
 		/* Ensure packet is for us */
-		if (iface->hwlen <= sizeof(dhcp->chaddr) &&
-		    memcmp(dhcp->chaddr, iface->hwaddr, iface->hwlen))
+		if (ifp->hwlen <= sizeof(dhcp->chaddr) &&
+		    memcmp(dhcp->chaddr, ifp->hwaddr, ifp->hwlen))
 		{
 			char buf[sizeof(dhcp->chaddr) * 3];
 
 			syslog(LOG_DEBUG, "%s: xid 0x%x is not for hwaddr %s",
-			    iface->name, ntohl(dhcp->xid),
+			    ifp->name, ntohl(dhcp->xid),
 			    hwaddr_ntoa(dhcp->chaddr, sizeof(dhcp->chaddr),
 				buf, sizeof(buf)));
 			continue;
 		}
-		dhcp_handledhcp(iface, &dhcp, &from);
+		dhcp_handledhcp(ifp, &dhcp, &from);
 		if (state->raw_fd == -1)
 			break;
 	}
@@ -2610,31 +2658,38 @@ dhcp_handlepacket(void *arg)
 }
 
 static void
-dhcp_handleudp(void *arg)
+dhcp_handleudp1(struct dhcpcd_ctx *ctx, int *fd, const char *ifname)
 {
-	struct dhcpcd_ctx *ctx;
 	uint8_t buffer[sizeof(struct dhcp_message)];
 
-	ctx = arg;
-
 	/* Just read what's in the UDP fd and discard it as we always read
 	 * from the raw fd */
-	(void)read(ctx->udp_fd, buffer, sizeof(buffer));
+	if (read(*fd, buffer, sizeof(buffer)) == -1) {
+		syslog(LOG_ERR, "%s: %s: %m", ifname, __func__);
+		eloop_event_delete(ctx->eloop, *fd);
+		close(*fd);
+		*fd = -1;
+	}
+}
+
+static void
+dhcp_handleudp(void *arg)
+{
+	struct dhcpcd_ctx *ctx;
+
+	ctx = arg;
+	dhcp_handleudp1(arg, &ctx->udp_fd, NULL);
 }
 
 static void
 dhcp_handleifudp(void *arg)
 {
 	const struct interface *ifp;
-	const struct dhcp_state *state;
-	uint8_t buffer[sizeof(struct dhcp_message)];
+	struct dhcp_state *state;
 
 	ifp = arg;
-	state = D_CSTATE(ifp);
-
-	/* Just read what's in the UDP fd and discard it as we always read
-	 * from the raw fd */
-	(void)read(state->udp_fd, buffer, sizeof(buffer));
+	state = D_STATE(ifp);
+	dhcp_handleudp1(ifp->ctx, &state->udp_fd, ifp->name);
 }
 
 static int
@@ -2652,7 +2707,7 @@ dhcp_open(struct interface *ifp)
 
 	state = D_STATE(ifp);
 	if (state->raw_fd == -1) {
-		state->raw_fd = ipv4_opensocket(ifp, ETHERTYPE_IP);
+		state->raw_fd = if_openrawsocket(ifp, ETHERTYPE_IP);
 		if (state->raw_fd == -1) {
 			syslog(LOG_ERR, "%s: %s: %m", __func__, ifp->name);
 			return -1;
@@ -2766,7 +2821,7 @@ dhcp_init(struct interface *ifp)
 {
 	struct dhcp_state *state;
 	const struct if_options *ifo;
-	size_t len;
+	uint8_t len;
 	char buf[(sizeof(ifo->clientid) - 1) * 3];
 
 	state = D_STATE(ifp);
@@ -2786,7 +2841,7 @@ dhcp_init(struct interface *ifp)
 	    LEASEFILE, ifp->name);
 
 	ifo = ifp->options;
-	/* We need to drop the leasefile so that start_interface
+	/* We need to drop the leasefile so that dhcp_start
 	 * doesn't load it. */
 	if (ifo->options & DHCPCD_REQUEST)
 		unlink(state->leasefile);
@@ -2795,27 +2850,28 @@ dhcp_init(struct interface *ifp)
 	state->clientid = NULL;
 
 	if (*ifo->clientid) {
-		state->clientid = malloc(ifo->clientid[0] + 1);
+		state->clientid = malloc((size_t)(ifo->clientid[0] + 1));
 		if (state->clientid == NULL)
 			goto eexit;
-		memcpy(state->clientid, ifo->clientid, ifo->clientid[0] + 1);
+		memcpy(state->clientid, ifo->clientid,
+		    (size_t)(ifo->clientid[0]) + 1);
 	} else if (ifo->options & DHCPCD_CLIENTID) {
 		if (ifo->options & DHCPCD_DUID) {
 			state->clientid = malloc(ifp->ctx->duid_len + 6);
 			if (state->clientid == NULL)
 				goto eexit;
-			state->clientid[0] = ifp->ctx->duid_len + 5;
+			state->clientid[0] =(uint8_t)(ifp->ctx->duid_len + 5);
 			state->clientid[1] = 255; /* RFC 4361 */
 			memcpy(state->clientid + 2, ifo->iaid, 4);
 			memcpy(state->clientid + 6, ifp->ctx->duid,
 			    ifp->ctx->duid_len);
 		} else {
-			len = ifp->hwlen + 1;
-			state->clientid = malloc(len + 1);
+			len = (uint8_t)(ifp->hwlen + 1);
+			state->clientid = malloc((size_t)len + 1);
 			if (state->clientid == NULL)
 				goto eexit;
 			state->clientid[0] = len;
-			state->clientid[1] = ifp->family;
+			state->clientid[1] = (uint8_t)ifp->family;
 			memcpy(state->clientid + 2, ifp->hwaddr,
 			    ifp->hwlen);
 		}
@@ -2840,9 +2896,10 @@ eexit:
 	return -1;
 }
 
-void
-dhcp_start(struct interface *ifp)
+static void
+dhcp_start1(void *arg)
 {
+	struct interface *ifp = arg;
 	struct if_options *ifo = ifp->options;
 	struct dhcp_state *state;
 	struct stat st;
@@ -2923,7 +2980,7 @@ dhcp_start(struct interface *ifp)
 				state->offer = NULL;
 				state->lease.addr.s_addr = 0;
 			} else {
-				l = now.tv_sec - st.st_mtime;
+				l = (uint32_t)(now.tv_sec - st.st_mtime);
 				state->lease.leasetime -= l;
 				state->lease.renewaltime -= l;
 				state->lease.rebindtime -= l;
@@ -2950,6 +3007,24 @@ dhcp_start(struct interface *ifp)
 		dhcp_reboot(ifp);
 }
 
+void
+dhcp_start(struct interface *ifp)
+{
+	struct timeval tv;
+
+	if (!(ifp->options->options & DHCPCD_IPV4))
+		return;
+
+	tv.tv_sec = DHCP_MIN_DELAY;
+	tv.tv_usec = (suseconds_t)arc4random_uniform(
+	    (DHCP_MAX_DELAY - DHCP_MIN_DELAY) * 1000000);
+	timernorm(&tv);
+	syslog(LOG_DEBUG,
+	    "%s: delaying DHCP for %0.1f seconds",
+	    ifp->name, timeval_to_double(&tv));
+	eloop_timeout_add_tv(ifp->ctx->eloop, &tv, dhcp_start1, ifp);
+}
+
 void
 dhcp_handleifa(int type, struct interface *ifp,
 	const struct in_addr *addr,
@@ -2958,7 +3033,7 @@ dhcp_handleifa(int type, struct interface *ifp,
 {
 	struct dhcp_state *state;
 	struct if_options *ifo;
-	int i;
+	uint8_t i;
 
 	state = D_STATE(ifp);
 	if (state == NULL)
