@@ -30,6 +30,7 @@ const char dhcpcd_copyright[] = "Copyright (c) 2006-2014 Roy Marples";
 #define _WITH_DPRINTF /* Stop FreeBSD bitching */
 
 #include <sys/file.h>
+#include <sys/queue.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
 #include <sys/time.h>
@@ -38,6 +39,7 @@ const char dhcpcd_copyright[] = "Copyright (c) 2006-2014 Roy Marples";
 
 #include <ctype.h>
 #include <errno.h>
+#include <fcntl.h>
 #include <getopt.h>
 #include <limits.h>
 #include <paths.h>
@@ -58,17 +60,15 @@ const char dhcpcd_copyright[] = "Copyright (c) 2006-2014 Roy Marples";
 #include "dhcp6.h"
 #include "duid.h"
 #include "eloop.h"
+#include "if.h"
 #include "if-options.h"
-#include "if-pref.h"
 #include "ipv4.h"
 #include "ipv6.h"
 #include "ipv6nd.h"
-#include "net.h"
-#include "platform.h"
 #include "script.h"
 
 #ifdef USE_SIGNALS
-const int handle_sigs[] = {
+const int dhcpcd_handlesigs[] = {
 	SIGALRM,
 	SIGHUP,
 	SIGINT,
@@ -198,7 +198,7 @@ handle_exit_timeout(void *arg)
 			 * an infinite timeout. */
 			ctx->options &=
 			    ~(DHCPCD_WAITIP | DHCPCD_WAITIP4 | DHCPCD_WAITIP6);
-			daemonise(ctx);
+			dhcpcd_daemonise(ctx);
 		} else
 			eloop_exit(ctx->eloop, EXIT_FAILURE);
 		return;
@@ -211,7 +211,7 @@ handle_exit_timeout(void *arg)
 
 /* Returns the pid of the child, otherwise 0. */
 pid_t
-daemonise(struct dhcpcd_ctx *ctx)
+dhcpcd_daemonise(struct dhcpcd_ctx *ctx)
 {
 #ifdef THERE_IS_NO_FORK
 	eloop_timeout_delete(ctx->eloop, handle_exit_timeout, ctx);
@@ -292,20 +292,6 @@ daemonise(struct dhcpcd_ctx *ctx)
 #endif
 }
 
-struct interface *
-find_interface(struct dhcpcd_ctx *ctx, const char *ifname)
-{
-	struct interface *ifp;
-
-	if (ctx != NULL && ctx->ifaces != NULL) {
-		TAILQ_FOREACH(ifp, ctx->ifaces, next) {
-			if (strcmp(ifp->name, ifname) == 0)
-				return ifp;
-		}
-	}
-	return NULL;
-}
-
 static void
 stop_interface(struct interface *ifp)
 {
@@ -315,15 +301,19 @@ stop_interface(struct interface *ifp)
 	syslog(LOG_INFO, "%s: removing interface", ifp->name);
 	ifp->options->options |= DHCPCD_STOPPING;
 
-	// Remove the interface from our list
-	TAILQ_REMOVE(ifp->ctx->ifaces, ifp, next);
 	dhcp6_drop(ifp, NULL);
 	ipv6nd_drop(ifp);
 	dhcp_drop(ifp, "STOP");
 	eloop_timeout_delete(ctx->eloop, NULL, ifp);
 	if (ifp->options->options & DHCPCD_DEPARTED)
 		script_runreason(ifp, "DEPARTED");
-	free_interface(ifp);
+	else
+		script_runreason(ifp, "STOPPED");
+
+	// Remove the interface from our list
+	TAILQ_REMOVE(ifp->ctx->ifaces, ifp, next);
+	if_free(ifp);
+
 	if (!(ctx->options & (DHCPCD_MASTER | DHCPCD_TEST)))
 		eloop_exit(ctx->eloop, EXIT_FAILURE);
 }
@@ -349,20 +339,23 @@ configure_interface1(struct interface *ifp)
 		ifo->options &= ~(DHCPCD_ARP | DHCPCD_IPV4LL);
 	if (!(ifp->flags & (IFF_POINTOPOINT | IFF_LOOPBACK | IFF_MULTICAST)))
 		ifo->options &= ~DHCPCD_IPV6RS;
-	if (ifo->options & DHCPCD_LINK && carrier_status(ifp) == LINK_UNKNOWN)
+	if (ifo->options & DHCPCD_LINK && if_carrier(ifp) == LINK_UNKNOWN)
 		ifo->options &= ~DHCPCD_LINK;
 
 	if (ifo->metric != -1)
-		ifp->metric = ifo->metric;
+		ifp->metric = (unsigned int)ifo->metric;
 
 	if (!(ifo->options & DHCPCD_IPV6))
 		ifo->options &= ~DHCPCD_IPV6RS;
 
+	if (ifo->options & DHCPCD_SLAACPRIVATE)
+		ifo->options |= DHCPCD_IPV6RA_OWN;
+
 	/* We want to disable kernel interface RA as early as possible. */
 	if (ifo->options & DHCPCD_IPV6RS) {
-		ra_global = check_ipv6(ifp->ctx, NULL,
+		ra_global = if_checkipv6(ifp->ctx, NULL,
 		    ifp->ctx->options & DHCPCD_IPV6RA_OWN ? 1 : 0);
-		ra_iface = check_ipv6(ifp->ctx, ifp->name,
+		ra_iface = if_checkipv6(ifp->ctx, ifp->name,
 		    ifp->options->options & DHCPCD_IPV6RA_OWN ? 1 : 0);
 		if (ra_global == -1 || ra_iface == -1)
 			ifo->options &= ~DHCPCD_IPV6RS;
@@ -456,11 +449,10 @@ configure_interface1(struct interface *ifp)
 	/* If we are not sending an authentication option, don't require it */
 	if (!(ifo->auth.options & DHCPCD_AUTH_SEND))
 		ifo->auth.options &= ~DHCPCD_AUTH_REQUIRE;
-
 }
 
 int
-select_profile(struct interface *ifp, const char *profile)
+dhcpcd_selectprofile(struct interface *ifp, const char *profile)
 {
 	struct if_options *ifo;
 
@@ -486,23 +478,23 @@ static void
 configure_interface(struct interface *ifp, int argc, char **argv)
 {
 
-	select_profile(ifp, NULL);
+	dhcpcd_selectprofile(ifp, NULL);
 	add_options(ifp->ctx, ifp->name, ifp->options, argc, argv);
 	configure_interface1(ifp);
 }
 
 void
-handle_carrier(struct dhcpcd_ctx *ctx, int carrier, int flags,
+dhcpcd_handlecarrier(struct dhcpcd_ctx *ctx, int carrier, unsigned int flags,
     const char *ifname)
 {
 	struct interface *ifp;
 
-	ifp = find_interface(ctx, ifname);
+	ifp = if_find(ctx, ifname);
 	if (ifp == NULL || !(ifp->options->options & DHCPCD_LINK))
 		return;
 
 	if (carrier == LINK_UNKNOWN)
-		carrier = carrier_status(ifp); /* will set ifp->flags */
+		carrier = if_carrier(ifp); /* will set ifp->flags */
 	else
 		ifp->flags = flags;
 
@@ -514,6 +506,7 @@ handle_carrier(struct dhcpcd_ctx *ctx, int carrier, int flags,
 			if (ifp->carrier == LINK_UP)
 				syslog(LOG_INFO, "%s: carrier lost", ifp->name);
 			ifp->carrier = LINK_DOWN;
+			script_runreason(ifp, "NOCARRIER");
 			dhcp6_drop(ifp, "EXPIRE6");
 			ipv6nd_drop(ifp);
 			/* Don't blindly delete our knowledge of LL addresses.
@@ -522,7 +515,6 @@ handle_carrier(struct dhcpcd_ctx *ctx, int carrier, int flags,
 			 * do nothing. */
 			ipv6_free_ll_callbacks(ifp);
 			dhcp_drop(ifp, "EXPIRE");
-			script_runreason(ifp, "NOCARRIER");
 		}
 	} else if (carrier == LINK_UP && ifp->flags & IFF_UP) {
 		if (ifp->carrier != LINK_UP) {
@@ -532,13 +524,13 @@ handle_carrier(struct dhcpcd_ctx *ctx, int carrier, int flags,
 			/* BSD does not emit RTM_NEWADDR or RTM_CHGADDR when the
 			 * hardware address changes so we have to go
 			 * through the disovery process to work it out. */
-			handle_interface(ctx, 0, ifp->name);
+			dhcpcd_handleinterface(ctx, 0, ifp->name);
 #endif
 			if (ifp->wireless)
-				getifssid(ifp->name, ifp->ssid);
+				if_getssid(ifp->name, ifp->ssid);
 			configure_interface(ifp, ctx->argc, ctx->argv);
 			script_runreason(ifp, "CARRIER");
-			start_interface(ifp);
+			dhcpcd_startinterface(ifp);
 		}
 	}
 }
@@ -570,16 +562,30 @@ warn_iaid_conflict(struct interface *ifp, uint8_t *iaid)
 }
 
 void
-start_interface(void *arg)
+dhcpcd_startinterface(void *arg)
 {
 	struct interface *ifp = arg;
 	struct if_options *ifo = ifp->options;
-	int nolease;
 	size_t i;
 	char buf[DUID_LEN * 3];
 
+	/* Add our link-local address before upping the interface
+	 * so our RFC7217 address beats the hwaddr based one.
+	 * This is also a safety check incase it was ripped out
+	 * from under us. */
+	if (ifo->options & DHCPCD_IPV6 && ipv6_start(ifp) == -1) {
+		syslog(LOG_ERR, "%s: ipv6_start: %m", ifp->name);
+		ifo->options &= DHCPCD_IPV6;
+	}
+
+	if (!(ifp->flags & IFF_UP) && if_carrier(ifp) != LINK_UNKNOWN) {
+		if (if_up(ifp) == -1)
+			syslog(LOG_ERR, "%s: if_up: %m",
+			    ifp->name);
+	}
+
 	if (ifp->carrier == LINK_UNKNOWN)
-		handle_carrier(ifp->ctx, LINK_UNKNOWN, 0, ifp->name);
+		dhcpcd_handlecarrier(ifp->ctx, LINK_UNKNOWN, 0, ifp->name);
 	if (ifp->carrier == LINK_DOWN) {
 		syslog(LOG_INFO, "%s: waiting for carrier", ifp->name);
 		return;
@@ -619,10 +625,13 @@ start_interface(void *arg)
 			ipv6nd_startrs(ifp);
 
 		if (!(ifo->options & DHCPCD_IPV6RS)) {
+			ssize_t nolease;
+
 			if (ifo->options & DHCPCD_IA_FORCED)
 				nolease = dhcp6_start(ifp, DH6S_INIT);
 			else {
-				nolease = dhcp6_find_delegates(ifp);
+				dhcp6_find_delegates(ifp);
+				nolease = 0;
 				/* Enabling the below doesn't really make
 				 * sense as there is currently no standard
 				 * to push routes via DHCPv6.
@@ -653,15 +662,18 @@ handle_link(void *arg)
 	struct dhcpcd_ctx *ctx;
 
 	ctx = arg;
-	if (manage_link(ctx) == -1 && errno != ENXIO && errno != ENODEV)
-		syslog(LOG_ERR, "manage_link: %m");
+	if (if_managelink(ctx) == -1) {
+		syslog(LOG_ERR, "if_managelink: %m");
+		eloop_event_delete(ctx->eloop, ctx->link_fd);
+		close(ctx->link_fd);
+		ctx->link_fd = -1;
+	}
 }
 
 static void
 init_state(struct interface *ifp, int argc, char **argv)
 {
 	struct if_options *ifo;
-	const char *reason;
 
 	configure_interface(ifp, argc, argv);
 	ifo = ifp->options;
@@ -675,9 +687,24 @@ init_state(struct interface *ifp, int argc, char **argv)
 		ifo->options &= ~DHCPCD_IPV6RS;
 	}
 
+	/* Add our link-local address before upping the interface
+	 * so our RFC7217 address beats the hwaddr based one.
+	 * This needs to happen before PREINIT incase a hook script
+	 * inadvertently ups the interface. */
+	if (ifo->options & DHCPCD_IPV6 && ipv6_start(ifp) == -1) {
+		syslog(LOG_ERR, "%s: ipv6_start: %m", ifp->name);
+		ifo->options &= DHCPCD_IPV6;
+	}
+}
+
+static void
+run_preinit(struct interface *ifp)
+{
+	const char *reason;
+
 	reason = NULL; /* appease gcc */
-	if (ifo->options & DHCPCD_LINK) {
-		switch (carrier_status(ifp)) {
+	if (ifp->options->options & DHCPCD_LINK) {
+		switch (if_carrier(ifp)) {
 		case LINK_DOWN:
 			ifp->carrier = LINK_DOWN;
 			reason = "NOCARRIER";
@@ -700,8 +727,8 @@ init_state(struct interface *ifp, int argc, char **argv)
 		script_runreason(ifp, reason);
 }
 
-void
-handle_interface(void *arg, int action, const char *ifname)
+int
+dhcpcd_handleinterface(void *arg, int action, const char *ifname)
 {
 	struct dhcpcd_ctx *ctx;
 	struct if_head *ifs;
@@ -711,29 +738,33 @@ handle_interface(void *arg, int action, const char *ifname)
 
 	ctx = arg;
 	if (action == -1) {
-		ifp = find_interface(ctx, ifname);
-		if (ifp != NULL) {
-			ifp->options->options |= DHCPCD_DEPARTED;
-			stop_interface(ifp);
+		ifp = if_find(ctx, ifname);
+		if (ifp == NULL) {
+			errno = ESRCH;
+			return -1;
 		}
-		return;
+		ifp->options->options |= DHCPCD_DEPARTED;
+		stop_interface(ifp);
+		return 0;
 	}
 
 	/* If running off an interface list, check it's in it. */
-	if (ctx->ifc) {
+	if (ctx->ifc && action != 2) {
 		for (i = 0; i < ctx->ifc; i++)
 			if (strcmp(ctx->ifv[i], ifname) == 0)
 				break;
 		if (i >= ctx->ifc)
-			return;
+			return 0;
 	}
 
-	ifs = discover_interfaces(ctx, -1, UNCONST(argv));
+	i = -1;
+	ifs = if_discover(ctx, -1, UNCONST(argv));
 	TAILQ_FOREACH_SAFE(ifp, ifs, next, ifn) {
 		if (strcmp(ifp->name, ifname) != 0)
 			continue;
+		i = 0;
 		/* Check if we already have the interface */
-		ifl = find_interface(ctx, ifp->name);
+		ifl = if_find(ctx, ifp->name);
 		if (ifl) {
 			/* The flags and hwaddr could have changed */
 			ifl->flags = ifp->flags;
@@ -744,28 +775,33 @@ handle_interface(void *arg, int action, const char *ifname)
 			TAILQ_REMOVE(ifs, ifp, next);
 			TAILQ_INSERT_TAIL(ctx->ifaces, ifp, next);
 		}
-		if (action == 1) {
+		if (action > 0) {
 			init_state(ifp, ctx->argc, ctx->argv);
-			start_interface(ifp);
+			run_preinit(ifp);
+			dhcpcd_startinterface(ifp);
 		}
 	}
 
 	/* Free our discovered list */
 	while ((ifp = TAILQ_FIRST(ifs))) {
 		TAILQ_REMOVE(ifs, ifp, next);
-		free_interface(ifp);
+		if_free(ifp);
 	}
 	free(ifs);
+
+	if (i == -1)
+		errno = ENOENT;
+	return i;
 }
 
 void
-handle_hwaddr(struct dhcpcd_ctx *ctx, const char *ifname,
-    const uint8_t *hwaddr, size_t hwlen)
+dhcpcd_handlehwaddr(struct dhcpcd_ctx *ctx, const char *ifname,
+    const uint8_t *hwaddr, uint8_t hwlen)
 {
 	struct interface *ifp;
 	char buf[sizeof(ifp->hwaddr) * 3];
 
-	ifp = find_interface(ctx, ifname);
+	ifp = if_find(ctx, ifname);
 	if (ifp == NULL)
 		return;
 
@@ -787,14 +823,14 @@ handle_hwaddr(struct dhcpcd_ctx *ctx, const char *ifname,
 static void
 if_reboot(struct interface *ifp, int argc, char **argv)
 {
-	int oldopts;
+	unsigned long long oldopts;
 
 	oldopts = ifp->options->options;
 	script_runreason(ifp, "RECONFIGURE");
 	configure_interface(ifp, argc, argv);
 	dhcp_reboot_newopts(ifp, oldopts);
 	dhcp6_reboot(ifp);
-	start_interface(ifp);
+	dhcpcd_startinterface(ifp);
 }
 
 static void
@@ -803,28 +839,29 @@ reconf_reboot(struct dhcpcd_ctx *ctx, int action, int argc, char **argv, int oi)
 	struct if_head *ifs;
 	struct interface *ifn, *ifp;
 
-	ifs = discover_interfaces(ctx, argc - oi, argv + oi);
+	ifs = if_discover(ctx, argc - oi, argv + oi);
 	if (ifs == NULL)
 		return;
 
 	while ((ifp = TAILQ_FIRST(ifs))) {
 		TAILQ_REMOVE(ifs, ifp, next);
-		ifn = find_interface(ctx, ifp->name);
+		ifn = if_find(ctx, ifp->name);
 		if (ifn) {
 			if (action)
 				if_reboot(ifn, argc, argv);
 			else
 				ipv4_applyaddr(ifn);
-			free_interface(ifp);
+			if_free(ifp);
 		} else {
-			init_state(ifp, argc, argv);
 			TAILQ_INSERT_TAIL(ctx->ifaces, ifp, next);
-			start_interface(ifp);
+			init_state(ifp, argc, argv);
+			run_preinit(ifp);
+			dhcpcd_startinterface(ifp);
 		}
 	}
 	free(ifs);
 
-	sort_interfaces(ctx);
+	ipv4_sortinterfaces(ctx);
 }
 
 static void
@@ -945,8 +982,8 @@ signal_init(void (*func)(int, siginfo_t *, void *), sigset_t *oldset)
 	sa.sa_flags = SA_SIGINFO;
 	sigemptyset(&sa.sa_mask);
 
-	for (i = 0; handle_sigs[i]; i++) {
-		if (sigaction(handle_sigs[i], &sa, NULL) == -1)
+	for (i = 0; dhcpcd_handlesigs[i]; i++) {
+		if (sigaction(dhcpcd_handlesigs[i], &sa, NULL) == -1)
 			return -1;
 	}
 	return 0;
@@ -954,13 +991,13 @@ signal_init(void (*func)(int, siginfo_t *, void *), sigset_t *oldset)
 #endif
 
 int
-handle_args(struct dhcpcd_ctx *ctx, struct fd_list *fd, int argc, char **argv)
+dhcpcd_handleargs(struct dhcpcd_ctx *ctx, struct fd_list *fd,
+    int argc, char **argv)
 {
 	struct interface *ifp;
 	int do_exit = 0, do_release = 0, do_reboot = 0;
 	int opt, oi = 0;
-	ssize_t len;
-	size_t l;
+	size_t len, l;
 	struct iovec iov[2];
 	char *tmp, *p;
 
@@ -993,16 +1030,21 @@ handle_args(struct dhcpcd_ctx *ctx, struct fd_list *fd, int argc, char **argv)
 			if (argc == 1) {
 				TAILQ_FOREACH(ifp, ctx->ifaces, next) {
 					len++;
+					if (D_STATE(ifp))
+						len++;
 					if (D6_STATE_RUNNING(ifp))
 						len++;
-					if (ipv6nd_has_ra(ifp))
+					if (ipv6nd_hasra(ifp))
 						len++;
 				}
-				len = write(fd->fd, &len, sizeof(len));
-				if (len != sizeof(len))
+				if (write(fd->fd, &len, sizeof(len)) !=
+				    sizeof(len))
 					return -1;
 				TAILQ_FOREACH(ifp, ctx->ifaces, next) {
-					send_interface(fd->fd, ifp);
+					if (send_interface(fd->fd, ifp) == -1)
+						syslog(LOG_ERR,
+						    "send_interface %d: %m",
+						    fd->fd);
 				}
 				return 0;
 			}
@@ -1011,21 +1053,25 @@ handle_args(struct dhcpcd_ctx *ctx, struct fd_list *fd, int argc, char **argv)
 				TAILQ_FOREACH(ifp, ctx->ifaces, next) {
 					if (strcmp(argv[opt], ifp->name) == 0) {
 						len++;
+						if (D_STATE(ifp))
+							len++;
 						if (D6_STATE_RUNNING(ifp))
 							len++;
-						if (ipv6nd_has_ra(ifp))
+						if (ipv6nd_hasra(ifp))
 							len++;
 					}
 				}
 			}
-			len = write(fd->fd, &len, sizeof(len));
-			if (len != sizeof(len))
+			if (write(fd->fd, &len, sizeof(len)) != sizeof(len))
 				return -1;
 			opt = 0;
 			while (argv[++opt] != NULL) {
 				TAILQ_FOREACH(ifp, ctx->ifaces, next) {
-					if (strcmp(argv[opt], ifp->name) == 0)
-						send_interface(fd->fd, ifp);
+					if (strcmp(argv[opt], ifp->name)== 0 &&
+					    send_interface(fd->fd, ifp) == -1)
+						syslog(LOG_ERR,
+						    "send_interface %d: %m",
+						    fd->fd);
 				}
 			}
 			return 0;
@@ -1082,7 +1128,7 @@ handle_args(struct dhcpcd_ctx *ctx, struct fd_list *fd, int argc, char **argv)
 			return 0;
 		}
 		for (oi = optind; oi < argc; oi++) {
-			if ((ifp = find_interface(ctx, argv[oi])) == NULL)
+			if ((ifp = if_find(ctx, argv[oi])) == NULL)
 				continue;
 			if (do_release) {
 				ifp->options->options |= DHCPCD_RELEASE;
@@ -1103,12 +1149,13 @@ int
 main(int argc, char **argv)
 {
 	struct dhcpcd_ctx ctx;
-	char pidfile[sizeof(PIDFILE) + IF_NAMESIZE];
+	char pidfile[sizeof(PIDFILE) + IF_NAMESIZE + 1];
 	struct if_options *ifo;
 	struct interface *ifp;
 	uint16_t family = 0;
 	int opt, oi = 0, i;
-	size_t len;
+	time_t t;
+	ssize_t len;
 #if defined(USE_SIGNALS) || !defined(THERE_IS_NO_FORK)
 	pid_t pid;
 #endif
@@ -1125,8 +1172,15 @@ main(int argc, char **argv)
 	siga = NULL;
 #endif
 	closefrom(3);
+#ifdef LOG_PERROR
 	openlog(PACKAGE, LOG_PERROR | LOG_PID, LOG_DAEMON);
+#else
+	openlog(PACKAGE, LOG_PID, LOG_DAEMON);
+#endif
 	setlogmask(LOG_UPTO(LOG_INFO));
+#ifndef LOG_PERROR
+	psyslog_prio = LOG_UPTO(LOG_INFO);
+#endif
 
 	/* Test for --help and --version */
 	if (argc > 1) {
@@ -1153,9 +1207,19 @@ main(int argc, char **argv)
 	{
 		switch (opt) {
 		case '4':
+			if (family) {
+				syslog(LOG_ERR, "cannot specify more than one"
+				    " address family");
+				goto exit_failure;
+			}
 			family = AF_INET;
 			break;
 		case '6':
+			if (family) {
+				syslog(LOG_ERR, "cannot specify more than one"
+				    " address family");
+				goto exit_failure;
+			}
 			family = AF_INET6;
 			break;
 		case 'f':
@@ -1237,8 +1301,12 @@ main(int argc, char **argv)
 	ctx.options &= ~DHCPCD_DAEMONISE;
 #endif
 
-	if (ctx.options & DHCPCD_DEBUG)
+	if (ctx.options & DHCPCD_DEBUG) {
 		setlogmask(LOG_UPTO(LOG_DEBUG));
+#ifndef LOG_PERROR
+		psyslog_prio = LOG_UPTO(LOG_DEBUG);
+#endif
+	}
 	if (ctx.options & DHCPCD_QUIET) {
 		i = open(_PATH_DEVNULL, O_RDWR);
 		if (i == -1)
@@ -1253,16 +1321,29 @@ main(int argc, char **argv)
 		/* If we have any other args, we should run as a single dhcpcd
 		 *  instance for that interface. */
 		if (optind == argc - 1 && !(ctx.options & DHCPCD_MASTER)) {
+			const char *per;
+
 			if (strlen(argv[optind]) > IF_NAMESIZE) {
 				syslog(LOG_ERR, "%s: interface name too long",
 				    argv[optind]);
 				goto exit_failure;
 			}
+			/* Allow a dhcpcd interface per address family */
+			switch(family) {
+			case AF_INET:
+				per = "-4";
+				break;
+			case AF_INET6:
+				per = "-6";
+				break;
+			default:
+				per = "";
+			}
 			snprintf(pidfile, sizeof(pidfile),
-			    PIDFILE, "-", argv[optind]);
+			    PIDFILE, "-", argv[optind], per);
 		}
 		else {
-			snprintf(pidfile, sizeof(pidfile), PIDFILE, "", "");
+			snprintf(pidfile, sizeof(pidfile), PIDFILE, "", "", "");
 			ctx.options |= DHCPCD_MASTER;
 		}
 	}
@@ -1370,12 +1451,15 @@ main(int argc, char **argv)
 		if (mkdir(DBDIR, 0755) == -1 && errno != EEXIST)
 			syslog(LOG_ERR, "mkdir `%s': %m", DBDIR);
 
-		ctx.pid_fd = open(pidfile,
-		    O_WRONLY | O_CREAT | O_CLOEXEC | O_NONBLOCK,
-		    0664);
+		opt = O_WRONLY | O_CREAT | O_NONBLOCK;
+#ifdef O_CLOEXEC
+		opt |= O_CLOEXEC;
+#endif
+		ctx.pid_fd = open(pidfile, opt, 0664);
 		if (ctx.pid_fd == -1)
 			syslog(LOG_ERR, "open `%s': %m", pidfile);
 		else {
+#ifdef LOCK_EX
 			/* Lock the file so that only one instance of dhcpcd
 			 * runs on an interface */
 			if (flock(ctx.pid_fd, LOCK_EX | LOCK_NB) == -1) {
@@ -1384,6 +1468,17 @@ main(int argc, char **argv)
 				ctx.pid_fd = -1;
 				goto exit_failure;
 			}
+#endif
+#ifndef O_CLOEXEC
+			if (fcntl(ctx.pid_fd, F_GETFD, &opt) == -1 ||
+			    fcntl(ctx.pid_fd, F_SETFD, opt | FD_CLOEXEC) == -1)
+			{
+				syslog(LOG_ERR, "fcntl: %m");
+				close(ctx.pid_fd);
+				ctx.pid_fd = -1;
+				goto exit_failure;
+			}
+#endif
 			write_pid(ctx.pid_fd, getpid());
 		}
 	}
@@ -1427,7 +1522,7 @@ main(int argc, char **argv)
 	 * instead.
 	 * We also need to open this before checking for interfaces below
 	 * so that we pickup any new addresses during the discover phase. */
-	ctx.link_fd = open_link_socket();
+	ctx.link_fd = if_openlinksocket();
 	if (ctx.link_fd == -1)
 		syslog(LOG_ERR, "open_link_socket: %m");
 	else
@@ -1439,9 +1534,9 @@ main(int argc, char **argv)
 	    (DHCPCD_MASTER | DHCPCD_DEV))
 		dev_start(&ctx);
 
-	ctx.ifaces = discover_interfaces(&ctx, ctx.ifc, ctx.ifv);
+	ctx.ifaces = if_discover(&ctx, ctx.ifc, ctx.ifv);
 	for (i = 0; i < ctx.ifc; i++) {
-		if (find_interface(&ctx, ctx.ifv[i]) == NULL)
+		if (if_find(&ctx, ctx.ifv[i]) == NULL)
 			syslog(LOG_ERR, "%s: interface not found or invalid",
 			    ctx.ifv[i]);
 	}
@@ -1457,62 +1552,47 @@ main(int argc, char **argv)
 		}
 	}
 
-	if (ctx.options & DHCPCD_BACKGROUND && daemonise(&ctx))
+	TAILQ_FOREACH(ifp, ctx.ifaces, next) {
+		init_state(ifp, argc, argv);
+	}
+
+	if (ctx.options & DHCPCD_BACKGROUND && dhcpcd_daemonise(&ctx))
 		goto exit_success;
 
 	opt = 0;
 	TAILQ_FOREACH(ifp, ctx.ifaces, next) {
-		init_state(ifp, argc, argv);
+		run_preinit(ifp);
 		if (ifp->carrier != LINK_DOWN)
 			opt = 1;
 	}
 
 	if (!(ctx.options & DHCPCD_BACKGROUND)) {
-		/* If we don't have a carrier, we may have to wait for a second
-		 * before one becomes available if we brought an interface up */
-		if (opt == 0 &&
-		    ctx.options & DHCPCD_LINK &&
-		    ctx.options & DHCPCD_WAITUP &&
-		    !(ctx.options & DHCPCD_WAITIP))
-		{
-			ts.tv_sec = 1;
-			ts.tv_nsec = 0;
-			nanosleep(&ts, NULL);
-			TAILQ_FOREACH(ifp, ctx.ifaces, next) {
-				handle_carrier(&ctx, LINK_UNKNOWN, 0,
-				    ifp->name);
-				if (ifp->carrier != LINK_DOWN) {
-					opt = 1;
-					break;
-				}
-			}
-		}
 		if (ctx.options & DHCPCD_MASTER)
-			i = ifo->timeout;
+			t = ifo->timeout;
 		else if ((ifp = TAILQ_FIRST(ctx.ifaces)))
-			i = ifp->options->timeout;
+			t = ifp->options->timeout;
 		else
-			i = 0;
+			t = 0;
 		if (opt == 0 &&
 		    ctx.options & DHCPCD_LINK &&
 		    !(ctx.options & DHCPCD_WAITIP))
 		{
 			syslog(LOG_WARNING, "no interfaces have a carrier");
-			if (daemonise(&ctx))
+			if (dhcpcd_daemonise(&ctx))
 				goto exit_success;
-		} else if (i > 0) {
+		} else if (t > 0) {
 			if (ctx.options & DHCPCD_IPV4LL)
 				ctx.options |= DHCPCD_TIMEOUT_IPV4LL;
-			eloop_timeout_add_sec(ctx.eloop, i,
+			eloop_timeout_add_sec(ctx.eloop, t,
 			    handle_exit_timeout, &ctx);
 		}
 	}
 	free_options(ifo);
 	ifo = NULL;
 
-	sort_interfaces(&ctx);
+	ipv4_sortinterfaces(&ctx);
 	TAILQ_FOREACH(ifp, ctx.ifaces, next) {
-		eloop_timeout_add_sec(ctx.eloop, 0, start_interface, ifp);
+		eloop_timeout_add_sec(ctx.eloop, 0, dhcpcd_startinterface, ifp);
 	}
 
 	i = eloop_start(&ctx);
@@ -1530,7 +1610,7 @@ exit1:
 	if (ctx.ifaces) {
 		while ((ifp = TAILQ_FIRST(ctx.ifaces))) {
 			TAILQ_REMOVE(ctx.ifaces, ifp, next);
-			free_interface(ifp);
+			if_free(ifp);
 		}
 		free(ctx.ifaces);
 	}
@@ -1542,7 +1622,7 @@ exit1:
 
 	free_options(ifo);
 	free_globals(&ctx);
-	restore_kernel_ra(&ctx);
+	if_rarestore(&ctx);
 	ipv4_ctxfree(&ctx);
 	ipv6_ctxfree(&ctx);
 	dev_stop(&ctx, !(ctx.options & DHCPCD_FORKED));
