@@ -32,9 +32,19 @@
 #include <sys/ioctl.h>
 #include <sys/param.h>
 
+#include <linux/if_addr.h>
+#include <linux/filter.h>
+#include <linux/if_packet.h>
 #include <linux/netlink.h>
 #include <linux/rtnetlink.h>
 
+#include <arpa/inet.h>
+#include <net/if.h>
+#include <netinet/if_ether.h>
+#include <netinet/in_systm.h>
+#include <netinet/in.h>
+#include <net/route.h>
+
 /* Support older kernels */
 #ifndef IFLA_WIRELESS
 # define IFLA_WIRELESS (IFLA_MASTER + 1)
@@ -48,30 +58,142 @@
 #endif
 
 #include <errno.h>
+#include <fcntl.h>
 #include <ctype.h>
 #include <stddef.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <syslog.h>
+#include <time.h>
 #include <unistd.h>
 
 #include "config.h"
 #include "common.h"
 #include "dev.h"
 #include "dhcp.h"
+#include "if.h"
 #include "ipv4.h"
 #include "ipv6.h"
-#include "net.h"
+#include "ipv6nd.h"
+
+#define bpf_insn		sock_filter
+#define BPF_SKIPTYPE
+#define BPF_ETHCOOK		-ETH_HLEN
+#define BPF_WHOLEPACKET	0x0fffffff /* work around buggy LPF filters */
+
+#include "bpf-filter.h"
+
+/* Broadcast address for IPoIB */
+static const uint8_t ipv4_bcast_addr[] = {
+	0x00, 0xff, 0xff, 0xff,
+	0xff, 0x12, 0x40, 0x1b, 0x00, 0x00, 0x00, 0x00,
+	0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff
+};
 
 #define PROC_INET6	"/proc/net/if_inet6"
 #define PROC_PROMOTE	"/proc/sys/net/ipv4/conf/%s/promote_secondaries"
 #define SYS_LAYER2	"/sys/class/net/%s/device/layer2"
 
+static const char *mproc =
+#if defined(__alpha__)
+	"system type"
+#elif defined(__arm__)
+	"Hardware"
+#elif defined(__avr32__)
+	"cpu family"
+#elif defined(__bfin__)
+	"BOARD Name"
+#elif defined(__cris__)
+	"cpu model"
+#elif defined(__frv__)
+	"System"
+#elif defined(__i386__) || defined(__x86_64__)
+	"vendor_id"
+#elif defined(__ia64__)
+	"vendor"
+#elif defined(__hppa__)
+	"model"
+#elif defined(__m68k__)
+	"MMU"
+#elif defined(__mips__)
+	"system type"
+#elif defined(__powerpc__) || defined(__powerpc64__)
+	"machine"
+#elif defined(__s390__) || defined(__s390x__)
+	"Manufacturer"
+#elif defined(__sh__)
+	"machine"
+#elif defined(sparc) || defined(__sparc__)
+	"cpu"
+#elif defined(__vax__)
+	"cpu"
+#else
+	NULL
+#endif
+	;
+
+int
+if_machinearch(char *str, size_t len)
+{
+	FILE *fp;
+	char buf[256];
+
+	if (mproc == NULL) {
+		errno = EINVAL;
+		return -1;
+	}
+
+	fp = fopen("/proc/cpuinfo", "r");
+	if (fp == NULL)
+		return -1;
+
+	while (fscanf(fp, "%255s : ", buf) != EOF) {
+		if (strncmp(buf, mproc, strlen(mproc)) == 0 &&
+		    fscanf(fp, "%255s", buf) == 1)
+		{
+		        fclose(fp);
+			return snprintf(str, len, ":%s", buf);
+		}
+	}
+	fclose(fp);
+	errno = ESRCH;
+	return -1;
+}
+
+static int
+check_proc_int(const char *path)
+{
+	FILE *fp;
+	int i;
+
+	fp = fopen(path, "r");
+	if (fp == NULL)
+		return -1;
+	if (fscanf(fp, "%d", &i) != 1)
+		i = -1;
+	fclose(fp);
+	return i;
+}
+
+static ssize_t
+write_path(const char *path, const char *val)
+{
+	FILE *fp;
+	ssize_t r;
+
+	fp = fopen(path, "w");
+	if (fp == NULL)
+		return -1;
+	r = fprintf(fp, "%s\n", val);
+	fclose(fp);
+	return r;
+}
+
 int
 if_init(struct interface *ifp)
 {
 	char path[sizeof(PROC_PROMOTE) + IF_NAMESIZE];
-	FILE *fp;
 	int n;
 
 	/* We enable promote_secondaries so that we can do this
@@ -82,29 +204,27 @@ if_init(struct interface *ifp)
 	 * This matches the behaviour of BSD which makes coding dhcpcd
 	 * a little easier as there's just one behaviour. */
 	snprintf(path, sizeof(path), PROC_PROMOTE, ifp->name);
-
-	fp = fopen(path, "w");
-	if (fp == NULL)
+	n = check_proc_int(path);
+	if (n == -1)
 		return errno == ENOENT ? 0 : -1;
-	n = fprintf(fp, "1");
-	fclose(fp);
-	return n == -1 ? -1 : 0;
+	if (n == 1)
+		return 0;
+	return write_path(path, "1") == -1 ? -1 : 0;
 }
 
 int
 if_conf(struct interface *ifp)
 {
-	char path[sizeof(SYS_LAYER2) + IF_NAMESIZE], buf[1];
-	FILE *fp;
+	char path[sizeof(SYS_LAYER2) + IF_NAMESIZE];
+	int n;
 
 	/* Some qeth setups require the use of the broadcast flag. */
 	snprintf(path, sizeof(path), SYS_LAYER2, ifp->name);
-	fp = fopen(path, "r");
-	if (fp == NULL)
+	n = check_proc_int(path);
+	if (n == -1)
 		return errno == ENOENT ? 0 : -1;
-	if (fgets(buf, sizeof(buf), fp) != NULL && buf[0] == '0')
+	if (n == 0)
 		ifp->options->options |= DHCPCD_BROADCAST;
-	fclose(fp);
 	return 0;
 }
 
@@ -121,9 +241,23 @@ _open_link_socket(struct sockaddr_nl *nl)
 {
 	int fd;
 
+#ifdef SOCK_CLOEXEC
 	fd = socket(AF_NETLINK, SOCK_RAW | SOCK_CLOEXEC, NETLINK_ROUTE);
 	if (fd == -1)
 		return -1;
+#else
+	int flags;
+
+	fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
+	if (fd == -1)
+		return -1;
+	if ((flags = fcntl(fd, F_GETFD, 0)) == -1 ||
+	    fcntl(fd, F_SETFD, flags | FD_CLOEXEC) == -1)
+	{
+		close(fd);
+	        return -1;
+	}
+#endif
 	nl->nl_family = AF_NETLINK;
 	if (bind(fd, (struct sockaddr *)nl, sizeof(*nl)) == -1)
 		return -1;
@@ -131,7 +265,7 @@ _open_link_socket(struct sockaddr_nl *nl)
 }
 
 int
-open_link_socket(void)
+if_openlinksocket(void)
 {
 	struct sockaddr_nl snl;
 
@@ -142,7 +276,7 @@ open_link_socket(void)
 	snl.nl_groups |= RTMGRP_IPV4_ROUTE | RTMGRP_IPV4_IFADDR;
 #endif
 #ifdef INET6
-	snl.nl_groups |= RTMGRP_IPV6_ROUTE | RTMGRP_IPV6_IFADDR;
+	snl.nl_groups |= RTMGRP_IPV6_ROUTE | RTMGRP_IPV6_IFADDR | RTMGRP_NEIGH;
 #endif
 
 	return _open_link_socket(&snl);
@@ -153,33 +287,30 @@ get_netlink(struct dhcpcd_ctx *ctx, int fd, int flags,
     int (*callback)(struct dhcpcd_ctx *, struct nlmsghdr *))
 {
 	char *buf = NULL, *nbuf;
-	ssize_t buflen = 0, bytes;
+	ssize_t bytes;
+	size_t buflen;
 	struct nlmsghdr *nlm;
 	struct sockaddr_nl nladdr;
 	socklen_t nladdr_len = sizeof(nladdr);
-	int r = -1;
+	int r;
 
+	buflen = 0;
+	r = -1;
 	for (;;) {
 		bytes = recv(fd, NULL, 0,
 		    flags | MSG_PEEK | MSG_DONTWAIT | MSG_TRUNC);
-		if (bytes == -1) {
-			if (errno == EAGAIN) {
-				r = 0;
-				goto eexit;
-			}
-			if (errno == EINTR)
-				continue;
+		if (bytes == -1 || bytes == 0)
 			goto eexit;
-		} else if (bytes == buflen) {
+		if ((size_t)bytes == buflen) {
 			/* Support kernels older than 2.6.22 */
 			if (bytes == 0)
 				bytes = 512;
 			else
 				bytes *= 2;
 		}
-		if (buflen < bytes) {
+		if (buflen < (size_t)bytes) {
 			/* Alloc 1 more so we work with older kernels */
-			buflen = bytes + 1;
+			buflen = (size_t)bytes + 1;
 			nbuf = realloc(buf, buflen);
 			if (nbuf == NULL)
 				goto eexit;
@@ -187,15 +318,8 @@ get_netlink(struct dhcpcd_ctx *ctx, int fd, int flags,
 		}
 		bytes = recvfrom(fd, buf, buflen, flags,
 		    (struct sockaddr *)&nladdr, &nladdr_len);
-		if (bytes == -1) {
-			if (errno == EAGAIN) {
-				r = 0;
-				goto eexit;
-			}
-			if (errno == EINTR)
-				continue;
+		if (bytes == -1 || bytes == 0)
 			goto eexit;
-		}
 
 		/* Check sender */
 		if (nladdr_len != sizeof(nladdr)) {
@@ -225,18 +349,18 @@ static int
 err_netlink(__unused struct dhcpcd_ctx *ctx, struct nlmsghdr *nlm)
 {
 	struct nlmsgerr *err;
-	int l;
+	size_t len;
 
 	if (nlm->nlmsg_type != NLMSG_ERROR)
 		return 0;
-	l = nlm->nlmsg_len - sizeof(*nlm);
-	if ((size_t)l < sizeof(*err)) {
+	len = nlm->nlmsg_len - sizeof(*nlm);
+	if (len < sizeof(*err)) {
 		errno = EBADMSG;
 		return -1;
 	}
 	err = (struct nlmsgerr *)NLMSG_DATA(nlm);
 	if (err->error == 0)
-		return l;
+		return (int)len;
 	errno = -err->error;
 	return -1;
 }
@@ -256,7 +380,8 @@ get_max_pid_t()
 static int
 link_route(struct dhcpcd_ctx *ctx, struct nlmsghdr *nlm)
 {
-	int len, idx, metric;
+	size_t len;
+	unsigned int idx, metric;
 	struct rtattr *rta;
 	struct rtmsg *rtm;
 	struct rt rt;
@@ -266,7 +391,7 @@ link_route(struct dhcpcd_ctx *ctx, struct nlmsghdr *nlm)
 		return 0;
 
 	len = nlm->nlmsg_len - sizeof(*nlm);
-	if ((size_t)len < sizeof(*rtm)) {
+	if (len < sizeof(*rtm)) {
 		errno = EBADMSG;
 		return -1;
 	}
@@ -300,12 +425,12 @@ link_route(struct dhcpcd_ctx *ctx, struct nlmsghdr *nlm)
 			    sizeof(rt.gate.s_addr));
 			break;
 		case RTA_OIF:
-			idx = *(int *)RTA_DATA(rta);
+			idx = *(unsigned int *)RTA_DATA(rta);
 			if (if_indextoname(idx, ifn))
-				rt.iface = find_interface(ctx, ifn);
+				rt.iface = if_find(ctx, ifn);
 			break;
 		case RTA_PRIORITY:
-			metric = *(int *)RTA_DATA(rta);
+			metric = *(unsigned int *)RTA_DATA(rta);
 			break;
 		}
 		rta = RTA_NEXT(rta, len);
@@ -324,7 +449,7 @@ link_route(struct dhcpcd_ctx *ctx, struct nlmsghdr *nlm)
 static int
 link_addr(struct dhcpcd_ctx *ctx, struct nlmsghdr *nlm)
 {
-	int len;
+	size_t len;
 	struct rtattr *rta;
 	struct ifaddrmsg *ifa;
 	char ifn[IF_NAMESIZE + 1];
@@ -340,14 +465,14 @@ link_addr(struct dhcpcd_ctx *ctx, struct nlmsghdr *nlm)
 		return 0;
 
 	len = nlm->nlmsg_len - sizeof(*nlm);
-	if ((size_t)len < sizeof(*ifa)) {
+	if (len < sizeof(*ifa)) {
 		errno = EBADMSG;
 		return -1;
 	}
 	ifa = NLMSG_DATA(nlm);
 	if (if_indextoname(ifa->ifa_index, ifn) == NULL)
 		return -1;
-	iface = find_interface(ctx, ifn);
+	iface = if_find(ctx, ifn);
 	if (iface == NULL)
 		return 1;
 	rta = (struct rtattr *) IFA_RTA(ifa);
@@ -397,7 +522,8 @@ link_addr(struct dhcpcd_ctx *ctx, struct nlmsghdr *nlm)
 	return 1;
 }
 
-static short l2addr_len(unsigned short if_type)
+static uint8_t
+l2addr_len(unsigned short if_type)
 {
 
 	switch (if_type) {
@@ -409,9 +535,10 @@ static short l2addr_len(unsigned short if_type)
 		return 8;
 	case ARPHRD_INFINIBAND:
 		return 20;
-	default:
-		return -1;
 	}
+
+	/* Impossible */
+	return 0;
 }
 
 static int
@@ -421,31 +548,81 @@ handle_rename(struct dhcpcd_ctx *ctx, unsigned int ifindex, const char *ifname)
 
 	TAILQ_FOREACH(ifp, ctx->ifaces, next) {
 		if (ifp->index == ifindex && strcmp(ifp->name, ifname)) {
-			handle_interface(ctx, -1, ifp->name);
+			dhcpcd_handleinterface(ctx, -1, ifp->name);
 			/* Let dev announce the interface for renaming */
 			if (!dev_listening(ctx))
-				handle_interface(ctx, 1, ifname);
+				dhcpcd_handleinterface(ctx, 1, ifname);
 			return 1;
 		}
 	}
 	return 0;
 }
 
+#ifdef INET6
+static int
+link_neigh(struct dhcpcd_ctx *ctx, struct nlmsghdr *nlm)
+{
+	struct ndmsg *r;
+	struct rtattr *rta;
+	size_t len;
+	struct in6_addr addr6;
+	int flags;
+
+	if (nlm->nlmsg_type != RTM_NEWNEIGH && nlm->nlmsg_type != RTM_DELNEIGH)
+		return 0;
+	if (nlm->nlmsg_len < sizeof(*r))
+		return -1;
+
+	r = NLMSG_DATA(nlm);
+	rta = (struct rtattr *)RTM_RTA(r);
+	len = RTM_PAYLOAD(nlm);
+        if (r->ndm_family == AF_INET6) {
+		flags = 0;
+		if (r->ndm_flags & NTF_ROUTER)
+			flags |= IPV6ND_ROUTER;
+		if (nlm->nlmsg_type == RTM_NEWNEIGH &&
+		    r->ndm_state &
+		    (NUD_REACHABLE | NUD_STALE | NUD_DELAY | NUD_PROBE |
+		     NUD_PERMANENT))
+		        flags |= IPV6ND_REACHABLE;
+		memset(&addr6, 0, sizeof(addr6));
+		while (RTA_OK(rta, len)) {
+			switch (rta->rta_type) {
+			case NDA_DST:
+				memcpy(&addr6.s6_addr, RTA_DATA(rta),
+				       sizeof(addr6.s6_addr));
+				break;
+			}
+			rta = RTA_NEXT(rta, len);
+		}
+		ipv6nd_neighbour(ctx, &addr6, flags);
+	}
+
+	return 1;
+}
+#endif
+
 static int
 link_netlink(struct dhcpcd_ctx *ctx, struct nlmsghdr *nlm)
 {
-	int len;
+	int r;
+	size_t len;
 	struct rtattr *rta, *hwaddr;
 	struct ifinfomsg *ifi;
 	char ifn[IF_NAMESIZE + 1];
 	struct interface *ifp;
 
-	len = link_route(ctx, nlm);
-	if (len != 0)
-		return len;
-	len = link_addr(ctx, nlm);
-	if (len != 0)
-		return len;
+	r = link_route(ctx, nlm);
+	if (r != 0)
+		return r;
+	r = link_addr(ctx, nlm);
+	if (r != 0)
+		return r;
+#ifdef INET6
+	r = link_neigh(ctx, nlm);
+	if (r != 0)
+		return r;
+#endif
 
 	if (nlm->nlmsg_type != RTM_NEWLINK && nlm->nlmsg_type != RTM_DELLINK)
 		return 0;
@@ -481,7 +658,7 @@ link_netlink(struct dhcpcd_ctx *ctx, struct nlmsghdr *nlm)
 	}
 
 	if (nlm->nlmsg_type == RTM_DELLINK) {
-		handle_interface(ctx, -1, ifn);
+		dhcpcd_handleinterface(ctx, -1, ifn);
 		return 1;
 	}
 
@@ -490,38 +667,41 @@ link_netlink(struct dhcpcd_ctx *ctx, struct nlmsghdr *nlm)
 	 * To trigger a valid hardware address pickup we need to pretend
 	 * that that don't exist until they have one. */
 	if (ifi->ifi_flags & IFF_MASTER && !hwaddr) {
-		handle_interface(ctx, -1, ifn);
+		dhcpcd_handleinterface(ctx, -1, ifn);
 		return 1;
 	}
 
 	/* Check for interface name change */
-	if (handle_rename(ctx, ifi->ifi_index, ifn))
+	if (handle_rename(ctx, (unsigned int)ifi->ifi_index, ifn))
 		    return 1;
 
 	/* Check for a new interface */
-	ifp = find_interface(ctx, ifn);
+	ifp = if_find(ctx, ifn);
 	if (ifp == NULL) {
 		/* If are listening to a dev manager, let that announce
 		 * the interface rather than the kernel. */
 		if (dev_listening(ctx) < 1)
-			handle_interface(ctx, 1, ifn);
+			dhcpcd_handleinterface(ctx, 1, ifn);
 		return 1;
 	}
 
 	/* Re-read hardware address and friends */
 	if (!(ifi->ifi_flags & IFF_UP) && hwaddr) {
-		len = l2addr_len(ifi->ifi_type);
-		if (hwaddr->rta_len == RTA_LENGTH(len))
-			handle_hwaddr(ctx, ifn, RTA_DATA(hwaddr), len);
+		uint8_t l;
+
+		l = l2addr_len(ifi->ifi_type);
+		if (hwaddr->rta_len == RTA_LENGTH(l))
+			dhcpcd_handlehwaddr(ctx, ifn, RTA_DATA(hwaddr), l);
 	}
 
-	handle_carrier(ctx, ifi->ifi_flags & IFF_RUNNING ? LINK_UP : LINK_DOWN,
+	dhcpcd_handlecarrier(ctx,
+	    ifi->ifi_flags & IFF_RUNNING ? LINK_UP : LINK_DOWN,
 	    ifi->ifi_flags, ifn);
 	return 1;
 }
 
 int
-manage_link(struct dhcpcd_ctx *ctx)
+if_managelink(struct dhcpcd_ctx *ctx)
 {
 
 	return get_netlink(ctx, ctx->link_fd, MSG_DONTWAIT, &link_netlink);
@@ -563,10 +743,10 @@ send_netlink(struct dhcpcd_ctx *ctx, struct nlmsghdr *hdr)
 	((struct rtattr *)(((ptrdiff_t)(nmsg))+NLMSG_ALIGN((nmsg)->nlmsg_len)))
 
 static int
-add_attr_l(struct nlmsghdr *n, unsigned int maxlen, int type,
-    const void *data, int alen)
+add_attr_l(struct nlmsghdr *n, unsigned short maxlen, unsigned short type,
+    const void *data, unsigned short alen)
 {
-	int len = RTA_LENGTH(alen);
+	unsigned short len = RTA_LENGTH(alen);
 	struct rtattr *rta;
 
 	if (NLMSG_ALIGN(n->nlmsg_len) + RTA_ALIGN(len) > maxlen) {
@@ -584,9 +764,10 @@ add_attr_l(struct nlmsghdr *n, unsigned int maxlen, int type,
 }
 
 static int
-add_attr_32(struct nlmsghdr *n, unsigned int maxlen, int type, uint32_t data)
+add_attr_32(struct nlmsghdr *n, unsigned short maxlen, unsigned short type,
+    uint32_t data)
 {
-	int len = RTA_LENGTH(sizeof(data));
+	unsigned short len = RTA_LENGTH(sizeof(data));
 	struct rtattr *rta;
 
 	if (NLMSG_ALIGN(n->nlmsg_len) + len > maxlen) {
@@ -618,6 +799,161 @@ struct nlmr
 };
 
 #ifdef INET
+int
+if_openrawsocket(struct interface *ifp, int protocol)
+{
+	int s;
+	union sockunion {
+		struct sockaddr sa;
+		struct sockaddr_in sin;
+		struct sockaddr_ll sll;
+		struct sockaddr_storage ss;
+	} su;
+	struct sock_fprog pf;
+#ifdef PACKET_AUXDATA
+	int n;
+#endif
+
+#ifdef SOCK_CLOEXEC
+	if ((s = socket(PF_PACKET, SOCK_DGRAM | SOCK_CLOEXEC | SOCK_NONBLOCK,
+	    htons(protocol))) == -1)
+		return -1;
+#else
+	int flags;
+
+	if ((s = socket(PF_PACKET, SOCK_DGRAM, htons(protocol))) == -1)
+		return -1;
+	if ((flags = fcntl(s, F_GETFD, 0)) == -1 ||
+	    fcntl(s, F_SETFD, flags | FD_CLOEXEC) == -1)
+	{
+		close(s);
+	        return -1;
+	}
+	if ((flags = fcntl(s, F_GETFL, 0)) == -1 ||
+	    fcntl(s, F_SETFL, flags | O_NONBLOCK) == -1)
+	{
+		close(s);
+	        return -1;
+	}
+#endif
+	/* Install the DHCP filter */
+	memset(&pf, 0, sizeof(pf));
+	if (protocol == ETHERTYPE_ARP) {
+		pf.filter = UNCONST(arp_bpf_filter);
+		pf.len = arp_bpf_filter_len;
+	} else {
+		pf.filter = UNCONST(dhcp_bpf_filter);
+		pf.len = dhcp_bpf_filter_len;
+	}
+	if (setsockopt(s, SOL_SOCKET, SO_ATTACH_FILTER, &pf, sizeof(pf)) != 0)
+		goto eexit;
+#ifdef PACKET_AUXDATA
+	n = 1;
+	if (setsockopt(s, SOL_PACKET, PACKET_AUXDATA, &n, sizeof(n)) != 0) {
+		if (errno != ENOPROTOOPT)
+			goto eexit;
+	}
+#endif
+
+	memset(&su, 0, sizeof(su));
+	su.sll.sll_family = PF_PACKET;
+	su.sll.sll_protocol = htons(protocol);
+	su.sll.sll_ifindex = (int)ifp->index;
+	if (bind(s, &su.sa, sizeof(su)) == -1)
+		goto eexit;
+	return s;
+
+eexit:
+	close(s);
+	return -1;
+}
+
+ssize_t
+if_sendrawpacket(const struct interface *ifp, int protocol,
+    const void *data, size_t len)
+{
+	const struct dhcp_state *state;
+	union sockunion {
+		struct sockaddr sa;
+		struct sockaddr_ll sll;
+		struct sockaddr_storage ss;
+	} su;
+	int fd;
+
+	memset(&su, 0, sizeof(su));
+	su.sll.sll_family = AF_PACKET;
+	su.sll.sll_protocol = htons(protocol);
+	su.sll.sll_ifindex = (int)ifp->index;
+	su.sll.sll_hatype = htons(ifp->family);
+	su.sll.sll_halen = (unsigned char)ifp->hwlen;
+	if (ifp->family == ARPHRD_INFINIBAND)
+		memcpy(&su.sll.sll_addr,
+		    &ipv4_bcast_addr, sizeof(ipv4_bcast_addr));
+	else
+		memset(&su.sll.sll_addr, 0xff, ifp->hwlen);
+	state = D_CSTATE(ifp);
+	if (protocol == ETHERTYPE_ARP)
+		fd = state->arp_fd;
+	else
+		fd = state->raw_fd;
+
+	return sendto(fd, data, len, 0, &su.sa, sizeof(su));
+}
+
+ssize_t
+if_readrawpacket(struct interface *ifp, int protocol,
+    void *data, size_t len, int *flags)
+{
+	struct iovec iov = {
+		.iov_base = data,
+		.iov_len = len,
+	};
+	struct msghdr msg = {
+		.msg_iov = &iov,
+		.msg_iovlen = 1,
+	};
+	struct dhcp_state *state;
+#ifdef PACKET_AUXDATA
+	unsigned char cmsgbuf[CMSG_LEN(sizeof(struct tpacket_auxdata))];
+	struct cmsghdr *cmsg;
+	struct tpacket_auxdata *aux;
+#endif
+
+	ssize_t bytes;
+	int fd = -1;
+
+#ifdef PACKET_AUXDATA
+	msg.msg_control = cmsgbuf;
+	msg.msg_controllen = sizeof(cmsgbuf);
+#endif
+
+	state = D_STATE(ifp);
+	if (protocol == ETHERTYPE_ARP)
+		fd = state->arp_fd;
+	else
+		fd = state->raw_fd;
+	bytes = recvmsg(fd, &msg, 0);
+	if (bytes == -1)
+		return -1;
+	*flags = RAW_EOF; /* We only ever read one packet */
+	if (bytes) {
+#ifdef PACKET_AUXDATA
+		for (cmsg = CMSG_FIRSTHDR(&msg);
+		     cmsg;
+		     cmsg = CMSG_NXTHDR(&msg, cmsg))
+		{
+			if (cmsg->cmsg_level == SOL_PACKET &&
+			    cmsg->cmsg_type == PACKET_AUXDATA) {
+				aux = (void *)CMSG_DATA(cmsg);
+				if (aux->tp_status & TP_STATUS_CSUMNOTREADY)
+					*flags |= RAW_PARTIALCSUM;
+			}
+		}
+#endif
+	}
+	return bytes;
+}
+
 int
 if_address(const struct interface *iface,
     const struct in_addr *address, const struct in_addr *netmask,
@@ -639,7 +975,7 @@ if_address(const struct interface *iface,
 	nlm.ifa.ifa_prefixlen = inet_ntocidr(*netmask);
 	/* This creates the aliased interface */
 	add_attr_l(&nlm.hdr, sizeof(nlm), IFA_LABEL,
-	    iface->name, strlen(iface->name) + 1);
+	    iface->name, (unsigned short)(strlen(iface->name) + 1));
 	add_attr_l(&nlm.hdr, sizeof(nlm), IFA_LOCAL,
 	    &address->s_addr, sizeof(address->s_addr));
 	if (action >= 0 && broadcast)
@@ -722,6 +1058,9 @@ if_address6(const struct ipv6_addr *ap, int action)
 	struct nlma nlm;
 	struct ifa_cacheinfo cinfo;
 	int retval = 0;
+#ifdef IFA_F_NOPREFIXROUTE
+	uint32_t flags;
+#endif
 
 	memset(&nlm, 0, sizeof(nlm));
 	nlm.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifaddrmsg));
@@ -736,7 +1075,7 @@ if_address6(const struct ipv6_addr *ap, int action)
 	nlm.ifa.ifa_prefixlen = ap->prefix_len;
 	/* This creates the aliased interface */
 	add_attr_l(&nlm.hdr, sizeof(nlm), IFA_LABEL,
-	    ap->iface->name, strlen(ap->iface->name) + 1);
+	    ap->iface->name, (unsigned short)(strlen(ap->iface->name) + 1));
 	add_attr_l(&nlm.hdr, sizeof(nlm), IFA_LOCAL,
 	    &ap->addr.s6_addr, sizeof(ap->addr.s6_addr));
 
@@ -748,16 +1087,23 @@ if_address6(const struct ipv6_addr *ap, int action)
 		    &cinfo, sizeof(cinfo));
 	}
 
+#ifdef IFA_F_NOPREFIXROUTE
+	if (!IN6_IS_ADDR_LINKLOCAL(&ap->addr)) {
+		flags = IFA_F_NOPREFIXROUTE;
+		add_attr_32(&nlm.hdr, sizeof(nlm), IFA_FLAGS, flags);
+	}
+#endif
+
 	if (send_netlink(ap->iface->ctx, &nlm.hdr) == -1)
 		retval = -1;
 	return retval;
 }
 
 static int
-rta_add_attr_32(struct rtattr *rta, unsigned int maxlen,
-    int type, uint32_t data)
+rta_add_attr_32(struct rtattr *rta, unsigned short maxlen,
+    unsigned short type, uint32_t data)
 {
-	unsigned int len = RTA_LENGTH(sizeof(data));
+	unsigned short len = RTA_LENGTH(sizeof(data));
 	struct rtattr *subrta;
 
 	if (RTA_ALIGN(rta->rta_len) + len > maxlen) {
@@ -769,7 +1115,7 @@ rta_add_attr_32(struct rtattr *rta, unsigned int maxlen,
 	subrta->rta_type = type;
 	subrta->rta_len = len;
 	memcpy(RTA_DATA(subrta), &data, sizeof(data));
-	rta->rta_len = NLMSG_ALIGN(rta->rta_len) + len;
+	rta->rta_len = (unsigned short)(NLMSG_ALIGN(rta->rta_len) + len);
 	return 0;
 }
 
@@ -804,21 +1150,24 @@ if_route6(const struct rt6 *rt, int action)
 			nlm.rt.rtm_scope = RT_SCOPE_LINK;
 		} else
 			nlm.rt.rtm_protocol = RTPROT_BOOT;
-		nlm.rt.rtm_type = RTN_UNICAST;
+		if (rt->flags & RTF_REJECT)
+			nlm.rt.rtm_type = RTN_UNREACHABLE;
+		else
+			nlm.rt.rtm_type = RTN_UNICAST;
 	}
 
 	nlm.rt.rtm_dst_len = ipv6_prefixlen(&rt->net);
 	add_attr_l(&nlm.hdr, sizeof(nlm), RTA_DST,
 	    &rt->dest.s6_addr, sizeof(rt->dest.s6_addr));
 
-	/* If destination == gateway then don't add the gateway */
-	if (!IN6_IS_ADDR_UNSPECIFIED(&rt->gate) &&
-	    !IN6_ARE_ADDR_EQUAL(&rt->dest, &rt->gate))
+	if (action >= 0 && !IN6_IS_ADDR_UNSPECIFIED(&rt->gate))
 		add_attr_l(&nlm.hdr, sizeof(nlm), RTA_GATEWAY,
 		    &rt->gate.s6_addr, sizeof(rt->gate.s6_addr));
 
-	add_attr_32(&nlm.hdr, sizeof(nlm), RTA_OIF, rt->iface->index);
-	add_attr_32(&nlm.hdr, sizeof(nlm), RTA_PRIORITY, rt->metric);
+	if (!(rt->flags & RTF_REJECT)) {
+		add_attr_32(&nlm.hdr, sizeof(nlm), RTA_OIF, rt->iface->index);
+		add_attr_32(&nlm.hdr, sizeof(nlm), RTA_PRIORITY, rt->metric);
+	}
 
 	if (rt->mtu) {
 		metrics->rta_type = RTA_METRICS;
@@ -834,7 +1183,7 @@ if_route6(const struct rt6 *rt, int action)
 }
 
 int
-in6_addr_flags(const char *ifname, const struct in6_addr *addr)
+if_addrflags6(const char *ifname, const struct in6_addr *addr)
 {
 	FILE *fp;
 	char *p, ifaddress[33], address[33], name[IF_NAMESIZE + 1];
@@ -871,4 +1220,91 @@ in6_addr_flags(const char *ifname, const struct in6_addr *addr)
 	errno = ESRCH;
 	return -1;
 }
+
+static const char *prefix = "/proc/sys/net/ipv6/conf";
+
+void
+if_rarestore(struct dhcpcd_ctx *ctx)
+{
+	char path[256];
+
+	for (; ctx->ra_restore_len > 0; ctx->ra_restore_len--) {
+		if (!(ctx->options & DHCPCD_FORKED)) {
+			syslog(LOG_DEBUG,
+			    "%s: restoring kernel IPv6 RA support",
+			    ctx->ra_restore[ctx->ra_restore_len - 1]);
+			snprintf(path, sizeof(path), "%s/%s/accept_ra",
+			    prefix, ctx->ra_restore[ctx->ra_restore_len - 1]);
+			if (write_path(path, "1") == -1 && errno != ENOENT)
+			    syslog(LOG_ERR, "write_path: %s: %m", path);
+		}
+		free(ctx->ra_restore[ctx->ra_restore_len - 1]);
+	}
+	free(ctx->ra_restore);
+	ctx->ra_restore = NULL;
+}
+
+int
+if_checkipv6(struct dhcpcd_ctx *ctx, const char *ifname, int own)
+{
+	int ra;
+	size_t i;
+	char path[256], *p, **nrest;
+
+	if (ifname == NULL)
+		ifname = "all";
+
+	snprintf(path, sizeof(path), "%s/%s/autoconf", prefix, ifname);
+	ra = check_proc_int(path);
+	if (ra != 1) {
+		if (!own) {
+			syslog(LOG_WARNING,
+			    "%s: IPv6 kernel autoconf disabled", ifname);
+			return -1;
+		}
+	} else if (ra != -1 && own) {
+		if (write_path(path, "0") == -1) {
+			syslog(LOG_ERR, "write_path: %s: %m", path);
+			return -1;
+		}
+	}
+
+	snprintf(path, sizeof(path), "%s/%s/accept_ra", prefix, ifname);
+	ra = check_proc_int(path);
+	if (ra == -1)
+		/* The sysctl probably doesn't exist, but this isn't an
+		 * error as such so just log it and continue */
+		syslog(errno == ENOENT ? LOG_DEBUG : LOG_WARNING,
+		    "%s: %m", path);
+	else if (ra != 0 && own) {
+		syslog(LOG_DEBUG, "%s: disabling kernel IPv6 RA support",
+		    ifname);
+		if (write_path(path, "0") == -1) {
+			syslog(LOG_ERR, "write_path: %s: %m", path);
+			return ra;
+		}
+		for (i = 0; i < ctx->ra_restore_len; i++)
+			if (strcmp(ctx->ra_restore[i], ifname) == 0)
+				break;
+		if (i == ctx->ra_restore_len) {
+			p = strdup(ifname);
+			if (p == NULL) {
+				syslog(LOG_ERR, "%s: %m", __func__);
+				return 0;
+			}
+			nrest = realloc(ctx->ra_restore,
+			    (ctx->ra_restore_len + 1) * sizeof(char *));
+			if (nrest == NULL) {
+				syslog(LOG_ERR, "%s: %m", __func__);
+				free(p);
+				return 0;
+			}
+			ctx->ra_restore = nrest;
+			ctx->ra_restore[ctx->ra_restore_len++] = p;
+		}
+		return 0;
+	}
+
+	return ra;
+}
 #endif
