@@ -51,6 +51,7 @@
 #include "dhcpcd.h"
 #include "dhcp6.h"
 #include "eloop.h"
+#include "if.h"
 #include "ipv6.h"
 #include "ipv6nd.h"
 #include "script.h"
@@ -116,11 +117,13 @@ struct nd_opt_dnssl {		/* DNSSL option RFC 6106 */
 #define IPV6_ADDR_INT16_MLL     0x02ff
 #endif
 
+#define ND6REACHABLE_TIMER	1
+
 /* Debugging Neighbor Solicitations is a lot of spam, so disable it */
 //#define DEBUG_NS
 //
 
-static void ipv6nd_handledata(void *arg);
+static void ipv6nd_handledata(void *);
 
 /*
  * Android ships buggy ICMP6 filter headers.
@@ -150,6 +153,14 @@ static void ipv6nd_handledata(void *arg);
 	memset(filterp, 0xff, sizeof(struct icmp6_filter));
 #endif
 
+/* Support older systems with different defines */
+#if !defined(IPV6_RECVHOPLIMIT) && defined(IPV6_HOPLIMIT)
+#define IPV6_RECVHOPLIMIT IPV6_HOPLIMIT
+#endif
+#if !defined(IPV6_RECVPKTINFO) && defined(IPV6_PKTINFO)
+#define IPV6_RECVPKTINFO IPV6_PKTINFO
+#endif
+
 static int
 ipv6nd_open(struct dhcpcd_ctx *dctx)
 {
@@ -161,12 +172,12 @@ ipv6nd_open(struct dhcpcd_ctx *dctx)
 	if (ctx->nd_fd != -1)
 		return ctx->nd_fd;
 #ifdef SOCK_CLOEXEC
-	ctx->nd_fd = socket(AF_INET6, SOCK_RAW | SOCK_CLOEXEC | SOCK_NONBLOCK,
+	ctx->nd_fd = socket(PF_INET6, SOCK_RAW | SOCK_CLOEXEC | SOCK_NONBLOCK,
 	    IPPROTO_ICMPV6);
 	if (ctx->nd_fd == -1)
 		return -1;
 #else
-	if ((ctx->nd_fd = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) == -1)
+	if ((ctx->nd_fd = socket(PF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) == -1)
 		return -1;
 	if ((on = fcntl(ctx->nd_fd, F_GETFD, 0)) == -1 ||
 	    fcntl(ctx->nd_fd, F_SETFD, on | FD_CLOEXEC) == -1)
@@ -305,6 +316,73 @@ ipv6nd_sendrsprobe(void *arg)
 		syslog(LOG_WARNING, "%s: no IPv6 Routers available", ifp->name);
 }
 
+static void
+ipv6nd_reachable(struct ra *rap, int flags)
+{
+
+	if (flags & IPV6ND_REACHABLE) {
+		if (rap->lifetime && rap->expired) {
+			syslog(LOG_INFO, "%s: %s is reachable again",
+			    rap->iface->name, rap->sfrom);
+			rap->expired = 0;
+			ipv6_buildroutes(rap->iface->ctx);
+			/* XXX Not really an RA */
+			script_runreason(rap->iface, "ROUTERADVERT");
+		}
+	} else {
+		/* Any error means it's really gone from the kernel
+		 * neighbour database */
+		if (rap->lifetime && !rap->expired) {
+			syslog(LOG_WARNING,
+			    "%s: %s is unreachable, expiring it",
+			    rap->iface->name, rap->sfrom);
+			rap->expired = 1;
+			ipv6_buildroutes(rap->iface->ctx);
+			/* XXX Not really an RA */
+			script_runreason(rap->iface, "ROUTERADVERT");
+		}
+	}
+}
+
+#ifdef HAVE_RTM_GETNEIGH
+void
+ipv6nd_neighbour(struct dhcpcd_ctx *ctx, struct in6_addr *addr, int flags)
+{
+	struct ra *rap;
+
+	if (ctx->ipv6) {
+	        TAILQ_FOREACH(rap, ctx->ipv6->ra_routers, next) {
+			if (IN6_ARE_ADDR_EQUAL(&rap->from, addr)) {
+				ipv6nd_reachable(rap, flags);
+				break;
+			}
+		}
+	}
+}
+
+#else
+
+static void
+ipv6nd_checkreachablerouters(void *arg)
+{
+	struct dhcpcd_ctx *ctx = arg;
+	struct ra *rap;
+	int flags;
+
+	TAILQ_FOREACH(rap, ctx->ipv6->ra_routers, next) {
+		flags = if_nd6reachable(rap->iface->name, &rap->from);
+		if (flags == -1) {
+			/* An error occured, so it's unreachable */
+			flags = 0;
+		}
+		ipv6nd_reachable(rap, flags);
+	}
+
+	eloop_timeout_add_sec(ctx->eloop, ND6REACHABLE_TIMER,
+	    ipv6nd_checkreachablerouters, ctx);
+}
+#endif
+
 static void
 ipv6nd_free_opts(struct ra *rap)
 {
@@ -344,11 +422,16 @@ void ipv6nd_freedrop_ra(struct ra *rap, int drop)
 	eloop_timeout_delete(rap->iface->ctx->eloop, NULL, rap);
 	if (!drop)
 		TAILQ_REMOVE(rap->iface->ctx->ipv6->ra_routers, rap, next);
+#ifndef HAVE_RTM_GETNEIGH
+	if (TAILQ_FIRST(rap->iface->ctx->ipv6->ra_routers) == NULL)
+		eloop_timeout_delete(rap->iface->ctx->eloop,
+		    ipv6nd_checkreachablerouters, rap->iface->ctx);
+#endif
 	ipv6_freedrop_addrs(&rap->addrs, drop, NULL);
 	ipv6nd_free_opts(rap);
 	free(rap->data);
-	free(rap->ns);
 	free(rap);
+
 }
 
 ssize_t
@@ -431,15 +514,17 @@ add_router(struct ipv6_ctx *ctx, struct ra *router)
 static int
 ipv6nd_scriptrun(struct ra *rap)
 {
-	int hasdns;
+	int hasdns, hasaddress, pid;
 	struct ipv6_addr *ap;
 	const struct ra_opt *rao;
 
+	hasaddress = 0;
 	/* If all addresses have completed DAD run the script */
 	TAILQ_FOREACH(ap, &rap->addrs, next) {
 		if ((ap->flags & (IPV6_AF_ONLINK | IPV6_AF_AUTOCONF)) ==
 		    (IPV6_AF_ONLINK | IPV6_AF_AUTOCONF))
 		{
+			hasaddress = 1;
 			if (!(ap->flags & IPV6_AF_DADCOMPLETED) &&
 			    ipv6_findaddr(ap->iface, &ap->addr))
 				ap->flags |= IPV6_AF_DADCOMPLETED;
@@ -470,8 +555,10 @@ ipv6nd_scriptrun(struct ra *rap)
 	}
 
 	script_runreason(rap->iface, "ROUTERADVERT");
-	if (hasdns)
-		hasdns = daemonise(rap->iface->ctx);
+	pid = 0;
+	if (hasdns && (hasaddress ||
+	    !(rap->flags & (ND_RA_FLAG_MANAGED | ND_RA_FLAG_OTHER))))
+		pid = dhcpcd_daemonise(rap->iface->ctx);
 #if 0
 	else if (options & DHCPCD_DAEMONISE &&
 	    !(options & DHCPCD_DAEMONISED) && new_data)
@@ -481,7 +568,15 @@ ipv6nd_scriptrun(struct ra *rap)
 		    ifp->name);
 }
 #endif
-	return hasdns;
+	return pid;
+}
+
+static void
+ipv6nd_addaddr(void *arg)
+{
+	struct ipv6_addr *ap = arg;
+
+	ipv6_addaddr(ap);
 }
 
 static void
@@ -491,17 +586,64 @@ ipv6nd_dadcallback(void *arg)
 	struct interface *ifp;
 	struct ra *rap;
 	int wascompleted, found;
+	struct timeval tv;
+	char buf[INET6_ADDRSTRLEN];
+	const char *p;
+	int dadcounter;
 
+	ifp = ap->iface;
 	wascompleted = (ap->flags & IPV6_AF_DADCOMPLETED);
 	ap->flags |= IPV6_AF_DADCOMPLETED;
-	if (ap->flags & IPV6_AF_DUPLICATED)
-		/* No idea what how to try and make another address :( */
+	if (ap->flags & IPV6_AF_DUPLICATED) {
+		ap->dadcounter++;
 		syslog(LOG_WARNING, "%s: DAD detected %s",
 		    ap->iface->name, ap->saddr);
 
-	if (!wascompleted) {
-		ifp = ap->iface;
+		/* Try and make another stable private address.
+		 * Because ap->dadcounter is always increamented,
+		 * a different address is generated. */
+		/* XXX Cache DAD counter per prefix/id/ssid? */
+		if (ifp->options->options & DHCPCD_SLAACPRIVATE &&
+		    ap->dadcounter < IDGEN_RETRIES)
+		{
+			syslog(LOG_INFO, "%s: deleting address %s",
+				ifp->name, ap->saddr);
+			if (if_deladdress6(ap) == -1 &&
+			    errno != EADDRNOTAVAIL && errno != ENXIO)
+				syslog(LOG_ERR, "if_deladdress6: %m");
+			dadcounter = ap->dadcounter;
+			if (ipv6_makestableprivate(&ap->addr,
+			    &ap->prefix, ap->prefix_len,
+			    ifp, &dadcounter) == -1)
+			{
+				syslog(LOG_ERR,
+				    "%s: ipv6_makestableprivate: %m",
+				    ifp->name);
+				return;
+			}
+			ap->dadcounter = dadcounter;
+			ap->flags &= ~(IPV6_AF_ADDED | IPV6_AF_DADCOMPLETED);
+			ap->flags |= IPV6_AF_NEW;
+			p = inet_ntop(AF_INET6, ap->addr.s6_addr,
+			    buf, sizeof(buf));
+			if (p)
+				snprintf(ap->saddr,
+				    sizeof(ap->saddr),
+				    "%s/%d",
+				    p, ap->prefix_len);
+			else
+				ap->saddr[0] = '\0';
+			tv.tv_sec = 0;
+			tv.tv_usec = (suseconds_t)arc4random_uniform(
+			    IDGEN_DELAY * USECINSEC);
+			timernorm(&tv);
+			eloop_timeout_add_tv(ifp->ctx->eloop, &tv,
+			    ipv6nd_addaddr, ap);
+			return;
+		}
+	}
 
+	if (!wascompleted) {
 		TAILQ_FOREACH(rap, ifp->ctx->ipv6->ra_routers, next) {
 			if (rap->iface != ifp)
 				continue;
@@ -531,9 +673,10 @@ ipv6nd_dadcallback(void *arg)
 
 static void
 ipv6nd_handlera(struct ipv6_ctx *ctx, struct interface *ifp,
-    struct icmp6_hdr *icp, ssize_t len)
+    struct icmp6_hdr *icp, size_t len)
 {
-	ssize_t l, m, n, olen;
+	size_t olen, l, m, n;
+	ssize_t r;
 	struct nd_router_advert *nd_ra;
 	struct nd_opt_prefix_info *pi;
 	struct nd_opt_mtu *mtu;
@@ -552,7 +695,7 @@ ipv6nd_handlera(struct ipv6_ctx *ctx, struct interface *ifp,
 	struct timeval expire;
 	uint8_t new_rap, new_data;
 
-	if ((size_t)len < sizeof(struct nd_router_advert)) {
+	if (len < sizeof(struct nd_router_advert)) {
 		syslog(LOG_ERR, "IPv6 RA packet too short from %s", ctx->sfrom);
 		return;
 	}
@@ -603,9 +746,6 @@ ipv6nd_handlera(struct ipv6_ctx *ctx, struct interface *ifp,
 		if (rap) {
 			free(rap->data);
 			rap->data_len = 0;
-			free(rap->ns);
-			rap->ns = NULL;
-			rap->nslen = 0;
 		}
 		new_data = 1;
 	} else
@@ -661,7 +801,7 @@ ipv6nd_handlera(struct ipv6_ctx *ctx, struct interface *ifp,
 	p = ((uint8_t *)icp) + sizeof(struct nd_router_advert);
 	lifetime = ~0U;
 	for (; len > 0; p += olen, len -= olen) {
-		if ((size_t)len < sizeof(struct nd_opt_hdr)) {
+		if (len < sizeof(struct nd_opt_hdr)) {
 			syslog(LOG_ERR, "%s: short option", ifp->name);
 			break;
 		}
@@ -682,29 +822,30 @@ ipv6nd_handlera(struct ipv6_ctx *ctx, struct interface *ifp,
 		case ND_OPT_PREFIX_INFORMATION:
 			pi = (struct nd_opt_prefix_info *)(void *)ndo;
 			if (pi->nd_opt_pi_len != 4) {
-				syslog(LOG_ERR,
+				syslog(new_data ? LOG_ERR : LOG_DEBUG,
 				    "%s: invalid option len for prefix",
 				    ifp->name);
-				break;
+				continue;
 			}
 			if (pi->nd_opt_pi_prefix_len > 128) {
-				syslog(LOG_ERR, "%s: invalid prefix len",
+				syslog(new_data ? LOG_ERR : LOG_DEBUG,
+				    "%s: invalid prefix len",
 				    ifp->name);
-				break;
+				continue;
 			}
 			if (IN6_IS_ADDR_MULTICAST(&pi->nd_opt_pi_prefix) ||
 			    IN6_IS_ADDR_LINKLOCAL(&pi->nd_opt_pi_prefix))
 			{
-				syslog(LOG_ERR,
+				syslog(new_data ? LOG_ERR : LOG_DEBUG,
 				    "%s: invalid prefix in RA", ifp->name);
-				break;
+				continue;
 			}
 			if (ntohl(pi->nd_opt_pi_preferred_time) >
 			    ntohl(pi->nd_opt_pi_valid_time))
 			{
-				syslog(LOG_ERR,
+				syslog(new_data ? LOG_ERR : LOG_DEBUG,
 				    "%s: pltime > vltime", ifp->name);
-				break;
+				continue;
 			}
 			TAILQ_FOREACH(ap, &rap->addrs, next)
 				if (ap->prefix_len ==pi->nd_opt_pi_prefix_len &&
@@ -719,10 +860,8 @@ ipv6nd_handlera(struct ipv6_ctx *ctx, struct interface *ifp,
 				    ND_OPT_PI_FLAG_ONLINK))
 					continue;
 				ap = calloc(1, sizeof(*ap));
-				if (ap == NULL) {
-					syslog(LOG_ERR, "%s: %m", __func__);
+				if (ap == NULL)
 					break;
-				}
 				ap->iface = rap->iface;
 				ap->flags = IPV6_AF_NEW;
 				ap->prefix_len = pi->nd_opt_pi_prefix_len;
@@ -733,9 +872,14 @@ ipv6nd_handlera(struct ipv6_ctx *ctx, struct interface *ifp,
 				    ND_OPT_PI_FLAG_AUTO)
 				{
 					ap->flags |= IPV6_AF_AUTOCONF;
-					ipv6_makeaddr(&ap->addr, ifp,
+					ap->dadcounter =
+					    ipv6_makeaddr(&ap->addr, ifp,
 					    &ap->prefix,
 					    pi->nd_opt_pi_prefix_len);
+					if (ap->dadcounter == -1) {
+						free(ap);
+						break;
+					}
 					cbp = inet_ntop(AF_INET6,
 					    ap->addr.s6_addr,
 					    buf, sizeof(buf));
@@ -769,6 +913,9 @@ ipv6nd_handlera(struct ipv6_ctx *ctx, struct interface *ifp,
 					opt = tmp;
 					opt[l - 1] = ' ';
 					strlcpy(opt + l, ap->saddr, m);
+				} else {
+					syslog(LOG_ERR, "%s: %m", __func__);
+					continue;
 				}
 			} else
 				opt = strdup(ap->saddr);
@@ -799,32 +946,32 @@ ipv6nd_handlera(struct ipv6_ctx *ctx, struct interface *ifp,
 			for (n = ndo->nd_opt_len - 1; n > 1; n -= 2,
 			    op += sizeof(addr.s6_addr))
 			{
-				m = ipv6_printaddr(NULL, 0, op, ifp->name);
-				if (m != -1)
-					l += m + 1;
+				r = ipv6_printaddr(NULL, 0, op, ifp->name);
+				if (r != -1)
+					l += (size_t)r + 1;
 			}
 			op = (uint8_t *)ndo;
 			op += offsetof(struct nd_opt_rdnss,
 			    nd_opt_rdnss_lifetime);
 			op += sizeof(rdnss->nd_opt_rdnss_lifetime);
 			tmp = opt = malloc(l);
-			if (opt) {
-				for (n = ndo->nd_opt_len - 1; n > 1; n -= 2,
-				    op += sizeof(addr.s6_addr))
-				{
-					m = ipv6_printaddr(tmp, l, op,
-					    ifp->name);
-					if (m != -1) {
-						l -= (m + 1);
-						tmp += m;
-						*tmp++ = ' ';
-					}
+			if (opt == NULL)
+				continue;
+			for (n = ndo->nd_opt_len - 1; n > 1; n -= 2,
+			    op += sizeof(addr.s6_addr))
+			{
+				r = ipv6_printaddr(tmp, l, op,
+				    ifp->name);
+				if (r != -1) {
+					l -= ((size_t)r + 1);
+					tmp += (size_t)r;
+					*tmp++ = ' ';
 				}
-				if (tmp != opt)
-					(*--tmp) = '\0';
-				else
-					*opt = '\0';
 			}
+			if (tmp != opt)
+				(*--tmp) = '\0';
+			else
+				*opt = '\0';
 			break;
 
 		case ND_OPT_DNSSL:
@@ -834,21 +981,27 @@ ipv6nd_handlera(struct ipv6_ctx *ctx, struct interface *ifp,
 			    nd_opt_dnssl_lifetime);
 			op += sizeof(dnssl->nd_opt_dnssl_lifetime);
 			n = (dnssl->nd_opt_dnssl_len - 1) * 8;
-			l = decode_rfc3397(NULL, 0, n, op);
-			if (l < 1) {
-				syslog(LOG_ERR, "%s: invalid DNSSL option",
+			r = decode_rfc3397(NULL, 0, op, n);
+			if (r < 1) {
+				syslog(new_data ? LOG_ERR : LOG_DEBUG,
+				    "%s: invalid DNSSL option",
 				    ifp->name);
+				continue;
 			} else {
+				l = (size_t)r;
 				tmp = malloc(l);
 				if (tmp) {
-					decode_rfc3397(tmp, l, n, op);
-					n = print_string(NULL, 0,
-					    l - 1, (const uint8_t *)tmp);
+					decode_rfc3397(tmp, l, op, n);
+					l -= 1;
+					n = (size_t)print_string(NULL, 0,
+					    (const uint8_t *)tmp, l);
 					opt = malloc(n);
 					if (opt)
 						print_string(opt, n,
-						    l - 1,
-						    (const uint8_t *)tmp);
+						    (const uint8_t *)tmp, l);
+					else
+						syslog(LOG_ERR, "%s: %m",
+						    __func__);
 					free(tmp);
 				}
 			}
@@ -862,12 +1015,13 @@ ipv6nd_handlera(struct ipv6_ctx *ctx, struct interface *ifp,
 			syslog(LOG_ERR, "%s: %m", __func__);
 			continue;
 		}
+
 		TAILQ_FOREACH(rao, &rap->options, next) {
 			if (rao->type == ndo->nd_opt_type &&
 			    strcmp(rao->option, opt) == 0)
 				break;
 		}
-		if (lifetime == 0) {
+		if (lifetime == 0 || *opt == '\0') {
 			if (rao) {
 				TAILQ_REMOVE(&rap->options, rao, next);
 				free(rao->option);
@@ -891,7 +1045,7 @@ ipv6nd_handlera(struct ipv6_ctx *ctx, struct interface *ifp,
 		if (lifetime == ~0U)
 			timerclear(&rao->expire);
 		else {
-			expire.tv_sec = lifetime;
+			expire.tv_sec = (time_t)lifetime;
 			expire.tv_usec = 0;
 			timeradd(&rap->received, &expire, &rao->expire);
 		}
@@ -911,16 +1065,6 @@ ipv6nd_handlera(struct ipv6_ctx *ctx, struct interface *ifp,
 	eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
 	eloop_timeout_delete(ifp->ctx->eloop, NULL, rap); /* reachable timer */
 
-	/* If we're owning the RA then we need to try and ensure the
-	 * router is actually reachable */
-	if (ifp->options->options & DHCPCD_IPV6RA_OWN ||
-	    ifp->options->options & DHCPCD_IPV6RA_OWN_DEFAULT)
-	{
-		rap->nsprobes = 0;
-		if (rap->lifetime)
-			ipv6nd_proberouter(rap);
-	}
-
 handle_flag:
 	if (rap->flags & ND_RA_FLAG_MANAGED) {
 		if (new_data && dhcp6_start(ifp, DH6S_INIT) == -1)
@@ -940,44 +1084,61 @@ handle_flag:
 
 	/* Expire should be called last as the rap object could be destroyed */
 	ipv6nd_expirera(ifp);
+
+#ifndef HAVE_RTM_GETNEIGH
+	/* Start our reachability tests now */
+	ipv6nd_checkreachablerouters(ifp->ctx);
+#endif
 }
 
 int
-ipv6nd_has_ra(const struct interface *ifp)
+ipv6nd_hasra(const struct interface *ifp)
 {
 	const struct ra *rap;
 
 	if (ifp->ctx->ipv6) {
 		TAILQ_FOREACH(rap, ifp->ctx->ipv6->ra_routers, next)
-			if (rap->iface == ifp)
+			if (rap->iface == ifp && !rap->expired)
 				return 1;
 	}
 	return 0;
 }
 
+int
+ipv6nd_hasradhcp(const struct interface *ifp)
+{
+	const struct ra *rap;
+
+	if (ifp->ctx->ipv6) {
+		TAILQ_FOREACH(rap, ifp->ctx->ipv6->ra_routers, next) {
+			if (rap->iface == ifp &&
+			    !rap->expired &&
+			    (rap->flags & (ND_RA_FLAG_MANAGED | ND_RA_FLAG_OTHER)))
+				return 1;
+		}
+	}
+	return 0;
+}
+
 ssize_t
 ipv6nd_env(char **env, const char *prefix, const struct interface *ifp)
 {
-	ssize_t l;
-	size_t len;
+	size_t i, l, len;
 	const struct ra *rap;
 	const struct ra_opt *rao;
-	int i;
 	char buffer[32];
 	const char *optn;
 	char **pref, **mtu, **rdnss, **dnssl, ***var, *new;
 
-	i = 0;
-	l = 0;
+	i = l = 0;
 	TAILQ_FOREACH(rap, ifp->ctx->ipv6->ra_routers, next) {
 		i++;
 		if (rap->iface != ifp)
 			continue;
 		if (env) {
 			snprintf(buffer, sizeof(buffer),
-			    "ra%d_from", i);
-			if (setvar(&env, prefix, buffer, rap->sfrom) == -1)
-				return -1;
+			    "ra%zu_from", i);
+			setvar(&env, prefix, buffer, rap->sfrom);
 		}
 		l++;
 
@@ -988,7 +1149,7 @@ ipv6nd_env(char **env, const char *prefix, const struct interface *ifp)
 			var = NULL;
 			switch(rao->type) {
 			case ND_OPT_PREFIX_INFORMATION:
-				optn = "prefix";
+				optn = "prefix"; /* really address */
 				var = &pref;
 				break;
 			case ND_OPT_MTU:
@@ -1019,7 +1180,7 @@ ipv6nd_env(char **env, const char *prefix, const struct interface *ifp)
 						continue;
 					} else
 						new++;
-					len = (new - **var) +
+					len = (size_t)(new - **var) +
 					    strlen(rao->option) + 1;
 					if (len > strlen(**var))
 						new = realloc(**var, len);
@@ -1029,7 +1190,9 @@ ipv6nd_env(char **env, const char *prefix, const struct interface *ifp)
 						**var = new;
 						new = strchr(**var, '=');
 						if (new) {
-							len -= (new - **var);
+							len -=
+							    (size_t)
+							    (new - **var);
 							strlcpy(new + 1,
 							    rao->option,
 							    len - 1);
@@ -1041,30 +1204,27 @@ ipv6nd_env(char **env, const char *prefix, const struct interface *ifp)
 				}
 				len = strlen(rao->option) + 1;
 				new = realloc(**var, strlen(**var) + 1 + len);
-				if (new == NULL)
-					return -1;
-				**var = new;
-				new += strlen(new);
-				*new++ = ' ';
-				strlcpy(new, rao->option, len);
+				if (new) {
+					**var = new;
+					new += strlen(new);
+					*new++ = ' ';
+					strlcpy(new, rao->option, len);
+				} else
+					syslog(LOG_ERR, "%s: %m", __func__);
 				continue;
 			}
 			if (env) {
 				snprintf(buffer, sizeof(buffer),
-				    "ra%d_%s", i, optn);
-				if (setvar(&env, prefix, buffer, rao->option)
-				    == -1)
-					return -1;
+				    "ra%zu_%s", i, optn);
+				setvar(&env, prefix, buffer, rao->option);
 			}
 		}
 	}
 
-	if (env) {
-		if (setvard(&env, prefix, "ra_count", i) == -1)
-			return -1;
-	}
+	if (env)
+		setvard(&env, prefix, "ra_count", i);
 	l++;
-	return l;
+	return (ssize_t)l;
 }
 
 void
@@ -1101,7 +1261,7 @@ ipv6nd_expirera(void *arg)
 			continue;
 		valid = 0;
 		if (rap->lifetime) {
-			lt.tv_sec = rap->lifetime;
+			lt.tv_sec = (time_t)rap->lifetime;
 			lt.tv_usec = 0;
 			timeradd(&rap->received, &lt, &expire);
 			if (rap->lifetime == 0 || timercmp(&now, &expire, >)) {
@@ -1110,7 +1270,6 @@ ipv6nd_expirera(void *arg)
 					    "%s: %s: router expired",
 					    ifp->name, rap->sfrom);
 					rap->expired = expired = 1;
-					ipv6nd_cancelproberouter(rap);
 				}
 			} else {
 				valid = 1;
@@ -1206,132 +1365,13 @@ ipv6nd_drop(struct interface *ifp)
 	}
 }
 
-static void
-ipv6nd_unreachable(void *arg)
-{
-	struct ra *rap = arg;
-	struct timeval tv;
-
-	/* We could add an unreachable flag and persist the information,
-	 * but that is more effort than it's probably worth. */
-	syslog(LOG_WARNING, "%s: %s is unreachable, expiring it",
-	    rap->iface->name, rap->sfrom);
-	rap->expired = 1;
-	ipv6_buildroutes(rap->iface->ctx);
-	script_runreason(rap->iface, "ROUTERADVERT"); /* XXX not RA */
-
-	/* We should still test if it's reachable or not so
-	 * incase it comes back to life and it's preferable. */
-	if (rap->reachable) {
-		ms_to_tv(&tv, rap->reachable);
-	} else {
-		tv.tv_sec = REACHABLE_TIME;
-		tv.tv_usec = 0;
-	}
-	eloop_timeout_add_tv(rap->iface->ctx->eloop,
-	    &tv, ipv6nd_proberouter, rap);
-}
-
-void
-ipv6nd_proberouter(void *arg)
-{
-	struct ra *rap = arg;
-	struct nd_neighbor_solicit *ns;
-	struct nd_opt_hdr *nd;
-	struct sockaddr_in6 dst;
-	struct cmsghdr *cm;
-	struct in6_pktinfo pi;
-	struct timeval tv, rtv;
-	struct ipv6_ctx *ctx;
-
-	if (ipv6nd_open(rap->iface->ctx) == -1) {
-		syslog(LOG_ERR, "%s: ipv6nd_open: %m", __func__);
-		return;
-	}
-
-	if (!rap->ns) {
-	        rap->nslen = sizeof(*ns) + ROUNDUP8(rap->iface->hwlen + 2);
-		rap->ns = calloc(1, rap->nslen);
-		if (rap->ns == NULL) {
-			syslog(LOG_ERR, "%s: %m", __func__);
-			return;
-		}
-		ns = (struct nd_neighbor_solicit *)(void *)rap->ns;
-		ns->nd_ns_type = ND_NEIGHBOR_SOLICIT;
-		//ns->nd_ns_cksum = 0;
-		//ns->nd_ns_code = 0;
-		//ns->nd_ns_reserved = 0;
-		ns->nd_ns_target = rap->from;
-		nd = (struct nd_opt_hdr *)(rap->ns + sizeof(*ns));
-		nd->nd_opt_type = ND_OPT_SOURCE_LINKADDR;
-		nd->nd_opt_len = (ROUNDUP8(rap->iface->hwlen + 2)) >> 3;
-		memcpy(nd + 1, rap->iface->hwaddr, rap->iface->hwlen);
-	}
-
-	memset(&dst, 0, sizeof(dst));
-	dst.sin6_family = AF_INET6;
-#ifdef SIN6_LEN
-	dst.sin6_len = sizeof(dst);
-#endif
-	memcpy(&dst.sin6_addr, &rap->from, sizeof(dst.sin6_addr));
-	dst.sin6_scope_id = rap->iface->index;
-
-	ctx = rap->iface->ctx->ipv6;
-	ctx->sndhdr.msg_name = (caddr_t)&dst;
-	ctx->sndhdr.msg_iov[0].iov_base = rap->ns;
-	ctx->sndhdr.msg_iov[0].iov_len = rap->nslen;
-
-	/* Set the outbound interface */
-	cm = CMSG_FIRSTHDR(&ctx->sndhdr);
-	if (cm == NULL) /* unlikely */
-		return;
-	cm->cmsg_level = IPPROTO_IPV6;
-	cm->cmsg_type = IPV6_PKTINFO;
-	cm->cmsg_len = CMSG_LEN(sizeof(pi));
-	memset(&pi, 0, sizeof(pi));
-	pi.ipi6_ifindex = rap->iface->index;
-	memcpy(CMSG_DATA(cm), &pi, sizeof(pi));
-
-#ifdef DEBUG_NS
-	syslog(LOG_INFO, "%s: sending IPv6 NS for %s",
-	    rap->iface->name, rap->sfrom);
-#endif
-	if (sendmsg(ctx->nd_fd, &ctx->sndhdr, 0) == -1) {
-		syslog(LOG_ERR, "%s: %s: sendmsg: %m",
-		    rap->iface->name, __func__);
-		return;
-	}
-
-	ms_to_tv(&tv, rap->retrans == 0 ? RETRANS_TIMER : rap->retrans);
-	ms_to_tv(&rtv, MIN_RANDOM_FACTOR);
-	timeradd(&tv, &rtv, &tv);
-	rtv.tv_sec = 0;
-	rtv.tv_usec = arc4random() % (MAX_RANDOM_FACTOR_U -MIN_RANDOM_FACTOR_U);
-	timeradd(&tv, &rtv, &tv);
-	eloop_timeout_add_tv(rap->iface->ctx->eloop,
-	    &tv, ipv6nd_proberouter, rap);
-
-	if (rap->nsprobes++ == 0)
-		eloop_timeout_add_sec(rap->iface->ctx->eloop,
-		    DELAY_FIRST_PROBE_TIME, ipv6nd_unreachable, rap);
-}
-
-void
-ipv6nd_cancelproberouter(struct ra *rap)
-{
-
-	eloop_timeout_delete(rap->iface->ctx->eloop, ipv6nd_proberouter, rap);
-	eloop_timeout_delete(rap->iface->ctx->eloop, ipv6nd_unreachable, rap);
-}
-
 static void
 ipv6nd_handlena(struct ipv6_ctx *ctx, struct interface *ifp,
-    struct icmp6_hdr *icp, ssize_t len)
+    struct icmp6_hdr *icp, size_t len)
 {
 	struct nd_neighbor_advert *nd_na;
 	struct ra *rap;
 	int is_router, is_solicited;
-	struct timeval tv;
 
 	if ((size_t)len < sizeof(struct nd_neighbor_advert)) {
 		syslog(LOG_ERR, "IPv6 NA packet too short from %s", ctx->sfrom);
@@ -1380,7 +1420,6 @@ ipv6nd_handlena(struct ipv6_ctx *ctx, struct interface *ifp,
 		syslog(LOG_INFO, "%s: %s is no longer a router",
 		    ifp->name, ctx->sfrom);
 		rap->expired = 1;
-		ipv6nd_cancelproberouter(rap);
 		ipv6_buildroutes(ifp->ctx);
 		script_runreason(ifp, "ROUTERADVERT");
 		return;
@@ -1394,17 +1433,6 @@ ipv6nd_handlena(struct ipv6_ctx *ctx, struct interface *ifp,
 			ipv6_buildroutes(ifp->ctx);
 			script_runreason(rap->iface, "ROUTERADVERT"); /* XXX */
 		}
-		rap->nsprobes = 0;
-		if (rap->reachable) {
-			ms_to_tv(&tv, rap->reachable);
-		} else {
-			tv.tv_sec = REACHABLE_TIME;
-			tv.tv_usec = 0;
-		}
-		eloop_timeout_add_tv(rap->iface->ctx->eloop,
-		    &tv, ipv6nd_proberouter, rap);
-		eloop_timeout_delete(rap->iface->ctx->eloop,
-		    ipv6nd_unreachable, rap);
 	}
 }
 
@@ -1425,8 +1453,11 @@ ipv6nd_handledata(void *arg)
 	ctx->rcvhdr.msg_controllen = CMSG_SPACE(sizeof(struct in6_pktinfo)) +
 	    CMSG_SPACE(sizeof(int));
 	len = recvmsg(ctx->nd_fd, &ctx->rcvhdr, 0);
-	if (len == -1) {
+	if (len == -1 || len == 0) {
 		syslog(LOG_ERR, "recvmsg: %m");
+		eloop_event_delete(dhcpcd_ctx->eloop, ctx->nd_fd);
+		close(ctx->nd_fd);
+		ctx->nd_fd = -1;
 		return;
 	}
 	ctx->sfrom = inet_ntop(AF_INET6, &ctx->from.sin6_addr,
@@ -1472,10 +1503,10 @@ ipv6nd_handledata(void *arg)
 	if (icp->icmp6_code == 0) {
 		switch(icp->icmp6_type) {
 			case ND_NEIGHBOR_ADVERT:
-				ipv6nd_handlena(ctx, ifp, icp, len);
+				ipv6nd_handlena(ctx, ifp, icp, (size_t)len);
 				return;
 			case ND_ROUTER_ADVERT:
-				ipv6nd_handlera(ctx, ifp, icp, len);
+				ipv6nd_handlera(ctx, ifp, icp, (size_t)len);
 				return;
 		}
 	}
@@ -1484,26 +1515,25 @@ ipv6nd_handledata(void *arg)
 	    icp->icmp6_type, icp->icmp6_code, ctx->sfrom);
 }
 
-int
-ipv6nd_startrs(struct interface *ifp)
+static void
+ipv6nd_startrs1(void *arg)
 {
+	struct interface *ifp = arg;
 	struct rs_state *state;
 
 	syslog(LOG_INFO, "%s: soliciting an IPv6 router", ifp->name);
 	if (ipv6nd_open(ifp->ctx) == -1) {
 		syslog(LOG_ERR, "%s: ipv6nd_open: %m", __func__);
-		return -1;
+		return;
 	}
 
-	eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
-
 	state = RS_STATE(ifp);
 	if (state == NULL) {
 		ifp->if_data[IF_DATA_IPV6ND] = calloc(1, sizeof(*state));
 		state = RS_STATE(ifp);
 		if (state == NULL) {
 			syslog(LOG_ERR, "%s: %m", __func__);
-			return -1;
+			return;
 		}
 	}
 
@@ -1512,10 +1542,26 @@ ipv6nd_startrs(struct interface *ifp)
 	ipv6nd_makersprobe(ifp);
 	if (state->rs == NULL) {
 		syslog(LOG_ERR, "%s: ipv6ns_makersprobe: %m", __func__);
-		return -1;
+		return;
 	}
 
 	state->rsprobes = 0;
 	ipv6nd_sendrsprobe(ifp);
-	return 0;
+}
+
+void
+ipv6nd_startrs(struct interface *ifp)
+{
+	struct timeval tv;
+
+	eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
+	tv.tv_sec = 0;
+	tv.tv_usec = (suseconds_t)arc4random_uniform(
+	    MAX_RTR_SOLICITATION_DELAY * 1000000);
+	timernorm(&tv);
+	syslog(LOG_DEBUG,
+	    "%s: delaying IPv6 router solictation for %0.1f seconds",
+	    ifp->name, timeval_to_double(&tv));
+	eloop_timeout_add_tv(ifp->ctx->eloop, &tv, ipv6nd_startrs1, ifp);
+	return;
 }
